import relationalai as rai

model = rai.Model("Test")

Person = model.Type("Person")
Criminal = model.Type("Criminal")

# loaded from `examples/data/people.csv`
with model.read("people.csv") as row:
    # id,name,age,account,criminal
    # P-00007333,Tommy Roman,18,,False

    person = Person.add(id=row.id).set(name=row.name, age=row.age)
    with row.account:
        person.set(account=row.account)
    with row.criminal:
        person.set(Criminal)

with model.query() as select:
    p = Person()
    res = select(p.id, p.name, p.age)

print(res.results)



#pyright: reportUnusedExpression=false
import relationalai as rai

graph = rai.Model("MyCoolDatabase")
Person = graph.Type("Person")
Adult = graph.Type("Adult")
Child = graph.Type("Woop")

with graph.rule():
    Person.add(name="Joe", age=74)
    Person.add(name="Bob", age=40)
    Person.add(name="Jane", age=10)

with graph.rule():
    p = Person()

with graph.rule():
    p = Person()
    with p.age >= 18:
        p.set(Adult)
    with p.age < 18:
        p.set(Child)

with graph.query() as select:
    # a = Adult();
    a = Person(Adult|Child)
    z = select(a, a.name, a.age)

print("People!\n")
print(z.results)









from gentest.emit import Document
import relationalai as rai
from relationalai.clients.test import Executor, Install, Query

graph = rai.clients.test.Graph("emit_playground")

Person = graph.Type("Person")
Adult = graph.Type("Adult")
Card = graph.Type("Card")

Person.add(name="Joe", age=54)
# Person.add(name="Bob", age=40)
# Person.add(name="Jane", age=10)

with graph.rule():
    p = Person()
    p.age
    p.set(Adult, name="Jeff")
    Card.add(name="Business", paper_color="ivory")
    # p.age > 18
    #p.set(Adult)


# with graph.query() as select:
#     p = Person();
#     res = select(p, p.name, p.age)



exec = graph._client
assert isinstance(exec, Executor)
print(str(exec))

doc = Document()

for block in exec.blocks:
    match block:
        case Install():
            doc.rule(block.task)
        case Query():
            doc.query(block.task)

h = "==[ Emit ]"
print(f"{h}{'='*(80-len(h))}")
print(doc.stringify())



import pytest

import relationalai as rai
from relationalai import std

model = rai.Model(name=globals().get("name", ""), config=globals().get("config"))
Point = model.Type("Point")

with model.rule():
    Point.add(x=1.1, y=-1, z=0)

# Check that the abs function works
with model.query(tag="abs") as select:
    p = Point(y=-1)
    abs1 = std.math.abs(p.y)
    abs2 = std.math.abs(-1.0)
    select(p, abs1, abs2)

# Check that the isclose function works
with model.query(tag="isclose") as select:
    p = Point(z=0)
    std.math.isclose(p.z, 1e-10)
    select(p, p.z)

# Check that the isclose function works with a tolerance
with model.query(tag="isclose_tolerance") as select:
    p = Point(z=0)
    std.math.isclose(p.z, 1e-8, 1e-9)
    select(p, p.z)  # Should return empty dataframe

# Check that the ceil function works
with model.query(tag="ceil") as select:
    p = Point(x=1.1)
    ceil1 = std.math.ceil(p.x)
    ceil2 = std.math.ceil(0.5)
    select(p, ceil1, ceil2)

# Check that the floor function works
with model.query(tag="floor") as select:
    p = Point(x=1.1)
    floor1 = std.math.floor(p.x)
    floor2 = std.math.floor(0.5)
    select(p, floor1, floor2)

# Check that the cbrt function works
with model.query(tag="cbrt") as select:
    p = Point(y=-1)
    cbrt1 = std.math.cbrt(p.y)
    cbrt2 = std.math.cbrt(8)
    select(p, cbrt1, cbrt2)

# Check that the sqrt function works
with model.query(tag="sqrt") as select:
    p = Point(z=0)
    sqrt1 = std.math.sqrt(p.z)
    sqrt2 = std.math.sqrt(4)
    select(p, sqrt1, sqrt2)

# Check that the sqrt function raises an error when given a negative number
with pytest.raises(ValueError):
    with model.query(tag="sqrt"):
        std.math.sqrt(-1)

# Check that the log function works
with model.query(tag="log") as select:
    p = Point(z=0)
    log1 = std.math.log(p.z)
    log2 = std.math.log(100, 10)
    select(p, log1, log2)

# Check that the log function raises an error when given a negative number
with pytest.raises(ValueError):
    with model.query(tag="log"):
        std.math.log(-1)

# Check that the sign function works
with model.query(tag="sign") as select:
    p = Point(x=1.1, y=-1)
    sign1 = std.math.sign(p.x)
    sign2 = std.math.sign(p.y)
    sign3 = std.math.sign(0)
    select(p, sign1, sign2, sign3)

# Check that the trunc_divide function works
with model.query(tag="trunc_divide") as select:
    p = Point(x=1.1)
    trunc_divide1 = std.math.trunc_divide(p.x, 2)
    trunc_divide2 = std.math.trunc_divide(-5, 2)
    select(p, trunc_divide1, trunc_divide2)

# Check that the sin, cos functions works
for func in [
    std.math.sin,
    std.math.cos,
    std.math.tan,
    std.math.acos,
    std.math.asin,
    std.math.atan,
    std.math.radians,
    std.math.degrees,
]:
    with model.query(tag=func.__name__) as select:
        p = Point()
        select(p, func(p.z), func(0))



import relationalai as rai
from relationalai.std import rel, aggregates

model = rai.Model(name=globals().get("name", "test_hector"), config=globals().get("config"))

DemandBundle = model.Type("DemandBundle")
Centroid = model.Type("Centroid")
AssignedCentroid = model.Type("AssignedCentroid")
DemandRegion = model.Type("DemandRegion")
Foo = model.Type("Foo")

with model.rule():
    DemandBundle.add(name="Bundle1").set(max_sources=10)

    c0 = Centroid.add(iter=0, cidx=0)
    c0.foo.extend([1, 2])
    c1 = Centroid.add(iter=0, cidx=1)
    c1.foo.add(1)

    DemandRegion.add(name=0,  demand=5, x=2)
    DemandRegion.add(name=1,  demand=8, x=2)
    DemandRegion.add(name=1,  demand=3, x=3)

with model.rule():
    bundles = DemandBundle()
    foo = Foo.add(name="foo")
    with model.scope():
        count_values = aggregates.count(bundles, bundles.max_sources, per=[bundles])
        count_rows = aggregates.count(bundles)
        v = aggregates.sum(bundles, bundles.max_sources, per=[bundles])
        with count_values == count_rows:
            foo.set(value=v)

with model.query() as select:
    tmp = Foo()
    response = select(tmp.name, tmp.value)

print(response.results)

with model.query() as select:
    c0 = Centroid(iter=0)
    new_x = aggregates.sum(c0.foo, per=[c0])
    new_d = rel.sqrt(new_x**2)
    res = select(c0.cidx, new_x, new_d)

print(res.results)

with model.query() as select:
    c0 = Centroid(iter=0)
    r = DemandRegion(name=c0.cidx)
    total_demand = aggregates.sum(r.demand, per=[c0])
    new_x = aggregates.sum(r, r.x * r.demand / total_demand, per=[c0])
    # 16/11 + 9/11 = 25/11 = 2.2727

    res = select(c0.cidx, new_x)

print(res.results)

with model.query() as select:
    c0 = Centroid(iter=0)
    r = DemandRegion(name=c0.cidx)
    total_demand = aggregates.sum(r.demand, per=[c0])
    new_x = aggregates.sum(r, r.x * r.demand / total_demand, per=[c0])
    # 16/11 + 9/11 = 25/11 = 2.2727

    res = select(c0.cidx, r.demand, new_x)

print(res.results)





# pyright: reportUnusedExpression=false
import relationalai as rai
from relationalai.std.graphs import Graph
from relationalai.std.aggregates import sum

#--------------------------------------------------
# Types
#--------------------------------------------------

model = rai.Model("MyCoolDatabase")
Person = model.Type("Person")
Transaction = model.Type("Transaction")

#--------------------------------------------------
# Load data
#--------------------------------------------------

with model.rule():
    joe = Person.add(name="Joe")
    jane = Person.add(name="Jane")
    bob = Person.add(name="Bob")
    dylan = Person.add(name="Dylan")
    susan = Person.add(name="Susan")

    Transaction.add(from_=joe, to=jane, amount=100)
    Transaction.add(from_=joe, to=jane, amount=1000)
    Transaction.add(from_=joe, to=jane, amount=10)
    Transaction.add(from_=joe, to=bob, amount=10)
    Transaction.add(from_=susan, to=dylan, amount=10)

#--------------------------------------------------
# Graph
#--------------------------------------------------

graph = Graph(model, weighted=True)

with model.rule():
    t = Transaction()
    # Aggregate the amounts of all txns between two people
    # @NOTE: Be careful not to specify multiple weights for a single edge, that can lead to no results!
    weight = sum(t, t.amount, per=[t.from_, t.to])
    graph.Edge.add(t.from_, t.to, label="Transfer", weight=weight)

with model.query() as select:
    p = Person()
    community_id = graph.compute.louvain(p)
    res = select(p.name, community_id)

print(res)



import relationalai as rai
from relationalai import std

# Query snapshots are output in same order as this list.
UNWEIGHTED_ALGOS = [
    "distance",
]

WEIGHTED_ALGOS = ["weighted_distance"]

UNARY_ALGOS = ["diameter_range"]

model = rai.Model(
    name=globals().get("name", "test_distance"), config=globals().get("config")
)
Object = model.Type("Object")
Distance = model.Type("Distance")
Range = model.Type("Range")

with model.rule():
    obj1 = Object.add(id=1)
    obj2 = Object.add(id=2)
    obj3 = Object.add(id=3)
    obj4 = Object.add(id=4)
    Distance.add(from_=obj1, to=obj2, weight=10)
    Distance.add(from_=obj2, to=obj3, weight=20)
    Distance.add(from_=obj3, to=obj3, weight=30)
    Distance.add(from_=obj2, to=obj4, weight=40)
    Distance.add(from_=obj4, to=obj3, weight=50)

undirected_graph = std.graphs.Graph(model, undirected=True)

with model.rule():
    d = Distance()
    undirected_graph.Edge.add(d.from_, d.to)

for algo_name in UNWEIGHTED_ALGOS:
    with model.query() as select:
        o1 = Object(id=1)
        o2 = Object(id=2)
        o3 = Object(id=3)
        o4 = Object(id=4)
        algo1 = getattr(undirected_graph.compute, algo_name)(o1, o3)
        algo2 = getattr(undirected_graph.compute, algo_name)(o2, o4)
        select(o1, o3, algo1)
        select(o2, o4, algo2)

weighted_graph = std.graphs.Graph(model, undirected=True, weighted=True)

with model.rule():
    d = Distance()
    weighted_graph.Edge.add(d.from_, d.to, weight=d.weight)

for algo_name in WEIGHTED_ALGOS:
    with model.query() as select:
        o1 = Object(id=1)
        o2 = Object(id=2)
        o3 = Object(id=3)
        o4 = Object(id=4)
        algo1 = getattr(weighted_graph.compute, algo_name)(o1, o3)
        algo2 = getattr(weighted_graph.compute, algo_name)(o2, o4)
        select(o1, o3, algo1)
        select(o2, o4, algo2)

with model.rule():
    obj1 = Object.add(id=1)
    obj2 = Object.add(id=2)
    obj3 = Object.add(id=3)
    obj4 = Object.add(id=4)
    Range.add(from_=obj1, to=obj2)
    Range.add(from_=obj3, to=obj2)
    Range.add(from_=obj4, to=obj3)

unary_graph = std.graphs.Graph(model, undirected=True)

with model.rule():
    r = Range()
    unary_graph.Edge.add(r.from_, r.to)

for algo_name in UNARY_ALGOS:
    with model.query() as select:
        algo = getattr(unary_graph.compute, algo_name)()
        response = select(algo)



#pyright: reportUnusedExpression=false
from typing import Tuple
import relationalai as rai
from relationalai.clients.snowflake import Snowflake
from relationalai.std import aggregates


def setup():
    import pandas as pd
    from pathlib import Path
    from snowflake.connector.pandas_tools import write_pandas
    model = rai.Model("cdc_simple", dry_run=False)
    model.resources._exec("create database if not exists sandbox")
    model.resources._exec("create schema if not exists sandbox.public")
    person = pd.read_csv(Path(__file__).parent / "data/people.csv")
    transaction = pd.read_csv(Path(__file__).parent / "data/transactions.csv")
    dfs = [person, transaction]
    names = ["PERSON", "TRANSACTION"]
    for df, name in zip(dfs, names):
        write_pandas(
            model.resources._conn,
            df,
            name,
            database="SANDBOX",
            schema="PUBLIC",
            auto_create_table=True
        )

model = rai.Model("cdc_simple", dry_run=False)
sf = Snowflake(model)
Person = sf.sandbox.public.person

with model.query() as select:
    person = Person()
    z = select(aggregates.count(person))

print(z.results)

@model.export("sandbox.public")
def person_id(name: str) -> Tuple[int]:
    p = Person()
    p.name == name
    return p.age,

if not model._client.dry_run:
    print("\nCalling person_id from Snowflake!\n")
    print(model.resources._exec("call sandbox.public.person_id('Tommy Roman');").fetch_pandas_all()) # pyright: ignore





from typing import Protocol
from hypothesis.core import HypothesisHandle

class HypothesisTestFn(Protocol):
    is_hypothesis_test: bool
    hypothesis: HypothesisHandle

    def __call__(self, *args, **kwargs):
        pass



import datetime
from _pytest.nodes import Node
import pytest

class FailInfo:
    def __init__(self, run: int, node: Node, error: Exception):
        self.run = run
        self.node = node
        self.error = error
        self.when = datetime.datetime.now()

class FailureCollectorPlugin:
    def __init__(self):
        self.failures: list[FailInfo] = []
        self.run_counter = 0

    @pytest.hookimpl(tryfirst=True)
    def pytest_sessionstart(self, session):
        self.failures.clear()
        self.run_counter += 1

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_makereport(self, item, call):
        if call.when == "call" and call.excinfo is not None:
            excinfo: pytest.ExceptionInfo = call.excinfo
            err = excinfo.value
            match err:
                case ExceptionGroup():
                    key_hash = getattr(err, 'key_hash', None)
                    for sub_err in err.exceptions:
                        if key_hash:
                            setattr(sub_err, 'key_hash', key_hash)

                        self.failures.append(FailInfo(self.run_counter, item, sub_err))
                case _:
                    self.failures.append(FailInfo(self.run_counter, item, err))

    def get_failures(self):
        return self.failures



from typing import Any
from ..dsl import Expression, build, get_graph
from ..metamodel import Builtins

#--------------------------------------------------
# Rel aggregations
#--------------------------------------------------

rank_desc_def = build.aggregate_def("reverse_sort")
rank_desc_def.parents.append(Builtins.Extender)
rank_asc_def = build.aggregate_def("sort")
rank_asc_def.parents.append(Builtins.Extender)

#--------------------------------------------------
# Aggregations
#--------------------------------------------------

def count(*args, per=[]) -> Any:
    return Expression(get_graph(), Builtins.count, [args, per])

def sum(*args, per=[]) -> Any:
    return Expression(get_graph(), Builtins.sum, [args, per])

def avg(*args, per=[]) -> Any:
    return Expression(get_graph(), Builtins.avg, [args, per])

def min(*args, per=[]) -> Any:
    return Expression(get_graph(), Builtins.min, [args, per])

def max(*args, per=[]) -> Any:
    return Expression(get_graph(), Builtins.max, [args, per])

def rank_asc(*args, per=[]) -> Any:
    return Expression(get_graph(), rank_asc_def, [args, per])

def rank_desc(*args, per=[]) -> Any:
    return Expression(get_graph(), rank_desc_def, [args, per])

def top(limit:int, *args, per=[]) -> Any:
    rank = rank_desc(*args, per=per)
    rank <= limit
    return rank

def bottom(limit:int, *args, per=[]) -> Any:
    rank = rank_asc(*args, per=per)
    rank <= limit
    return rank

#--------------------------------------------------
# Exports
#--------------------------------------------------

__all__ = ["count", "sum", "avg", "min", "max", "rank_asc", "rank_desc", "top", "bottom"]



import warnings
import utils
import en_core_web_sm
import requests

import relationalai as rai
from relationalai.std.graphs import Graph

nlp = en_core_web_sm.load()
warnings.filterwarnings("ignore")


def check_wiki_page(query):
    try:
        result = requests.get(f"https://en.wikipedia.org/wiki/{query}", verify=False)
    except Exception:
        pass

    if result.status_code == 200:
        return True
    return False


# --------------------------------------------------
# Load data
# --------------------------------------------------
articles = utils.get_articles(num_articles=20, use_local=True)

# --------------------------------------------------
# Types
# --------------------------------------------------
model = rai.Model("ArticleAnalysis")
Article = model.Type("Article")
ContentSource = model.Type("ContentSource")
Language = model.Type("Language")
Author = model.Type("Author")
People = model.Type("People")
Location = model.Type("Location")
Organization = model.Type("Organization")

# --------------------------------------------------
# Load data
# --------------------------------------------------
with model.rule(dynamic=True):
    for article in articles:
        source = ContentSource.add(name=article["source"]["title"])
        language = Language.add(name=article["lang"])
        author = Author.add(name=utils.get_author(article))

        a = Article.add(
            title=article.get("title", "Unknown Title"),
            trimmed_title=utils.trim_node_label(article.get("title", "Unknown Title")),
            sentiment=article["sentiment"] or "N/A",
            body=article["body"],
            dType=article["dataType"],
            published_on=article["dateTimePub"],
            date=article["date"],
            url=article["url"],
            language=language,
            author=author,
            source=source,
        )

        # get the nlp data for the article body
        text_nlp = nlp(article["body"])
        print("Processing entities in the articles using NLP...")

        # iterates over each entity recognized in the article.
        for ent in text_nlp.ents:
            print("Entity:", ent.text, "Label:", ent.label_)

            # explicitly skip entities containing the "%" symbol to prevent processing of numerical or statistical data, 
            # assuming such entities are less relevant to our context
            if "%" in ent.text:
                continue

            ent_text = ent.text.lower().replace("-", "").replace("'s", "")

            # checks if there's a Wikipedia page for the entity text. If not, the entity is skipped, using this as a criterion for notability and relevance.
            if not check_wiki_page(ent_text):
                continue

            # adds the entity to the Organization type and sets it as the organization of the article
            # if it's recognized as an organization.
            if ent.label_ == "ORG":
                organization = Organization.add(name=ent_text)
                a.organizations.add(organization)
            # adds the entity to the People type and sets it as the people of the article
            # if it's recognized as a person.
            elif ent.label_ == "PERSON":
                people = People.add(name=ent_text)
                a.peoples.add(people)
            # adds the entity to the Location type and sets it as the location of the article
            elif ent.label_ == "GPE":
                location = Location.add(name=ent_text)
                a.locations.add(location)

# --------------------------------------------------
# Query for articles to see what we have
# --------------------------------------------------
with model.query() as select:
    a = Article()
    response = select(a.title, a.body, a.date, a.source)
print(response.results)

# --------------------------------------------------
# Graph
# --------------------------------------------------
graph = Graph(model)
Node, Edge = graph.Node, graph.Edge

with model.rule():
    a = Article()
    Node.add(a, label=a.trimmed_title, type="article")
    Node.add(a.source, label=a.source.name, type="source")
    Node.add(a.language, label=a.language.name, type="language")
    Node.add(a.author, label=a.author.name, type="author")

    Node.add(a.peoples, label=a.peoples.name, type="people")
    Node.add(a.locations, label=a.locations.name, type="location")
    Node.add(a.organizations, label=a.organizations.name, type="organization")

    Edge.add(a, a.source, type="publishedBy")
    Edge.add(a, a.language, type="writtenIn")
    Edge.add(a, a.author, type="writtenBy")

    Edge.add(a, a.peoples, type="people")
    Edge.add(a, a.locations, type="location")
    Edge.add(a, a.organizations, type="organization")

graph.visualize(
    three=False,
    show_edge_label=True,
    node_label_size_factor=1.5,
    node_size_factor=1.5,
    node_hover_neighborhood=True,
    show_node_label=True,
    style={
        "node": {
            "size": utils.get_node_size_by_type,
            "color": utils.get_node_color_by_type,
        },
        "edge": {
            "opacity": 0.8,
            "color": utils.get_edge_color_by_type,
            "label": lambda t: f'{t["type"]}',
        },
    },
).display()



from io import StringIO
import sys
from typing import Any
from _pytest.nodes import Node
import pytest

class TestCollectorPlugin:
    def __init__(self):
        self.items: list[Node] = []

    def pytest_collection_modifyitems(self, items):
        for item in items:
            self.items.append(item)

_cached_tests: list[Node]|None = []

def collect_tests(force_recollect = False):
    global _cached_tests
    if force_recollect:
        _cached_tests = None
    if _cached_tests:
        return _cached_tests

    collector = TestCollectorPlugin()
    buf = StringIO()
    buf.isatty = lambda: True
    prev_stdout = sys.stdout
    sys.stdout = buf
    try:
        pytest.main(["--collect-only", "-q"], plugins=[collector])
    finally:
        sys.stdout = prev_stdout
        contents = buf.getvalue()
        if "ERROR" in contents:
            print("Failed to collect tests, see reason below:")
            print(contents)

    return collector.items

def pytest_node_fn(node: Node):
    n: Any = node
    assert callable(n._obj)
    return n._obj

def pytest_node_name(node: Node):
    n: Any = node
    return n.originalname

if __name__ == "__main__":
    items = collect_tests()
    print("COLLECTED")
    for item in items:
        fn = pytest_node_fn(item)
        print(item, getattr(fn, "__key_hash", None))



from abc import ABC
from dataclasses import dataclass, field
from typing import Sequence
from hypothesis import strategies as gen
from hypothesis.strategies import SearchStrategy as Strategy
import rich
from gentest.gen.error import BuilderConsumedError
from gentest.gen.group_limited import group_limiter, limited_by_group
from relationalai import metamodel as mm
from gentest.gen import ir
from gentest.gen.context import GenContext

def gen_python_ident(min_size=1, max_size=10):
    return gen.text(min_size=1, max_size=10, alphabet=gen.characters(
        whitelist_categories=('Lu', 'Ll', 'Lt', 'Lm', 'Lo', 'Nl'),
        whitelist_characters='_'))

#-------------------------------------------------------------------------------
# Constants
#-------------------------------------------------------------------------------

SUPPORTS_CONST_GEN = [mm.Builtins.Number, mm.Builtins.Int, mm.Builtins.Decimal, mm.Builtins.String, mm.Builtins.Bool, mm.Builtins.Any]

constant_int_strat = gen.builds(ir.Constant, type=gen.just(mm.Builtins.Int), value=gen.integers())
constant_decimal_strat = gen.builds(ir.Constant, type=gen.just(mm.Builtins.Decimal), value=gen.floats())
constant_number_strat = gen.builds(ir.Constant, type=gen.just(mm.Builtins.Number), value=gen.one_of(gen.integers(), gen.floats()))
constant_string_strat = gen.builds(ir.Constant, type=gen.just(mm.Builtins.String), value=gen.text())
constant_bool_strat = gen.builds(ir.Constant, type=gen.just(mm.Builtins.Bool), value=gen.booleans())
arbitrary_constant_strat = gen.one_of(
    constant_int_strat,
    constant_decimal_strat,
    constant_number_strat,
    constant_string_strat,
    constant_bool_strat)

def gen_constant(type: mm.Type|None):
    match type:
        case None | mm.Builtins.Any:
            return arbitrary_constant_strat
        case mm.Builtins.Int:
            return constant_int_strat
        case mm.Builtins.Decimal:
            return constant_decimal_strat
        case mm.Builtins.Number:
            return constant_number_strat
        case mm.Builtins.String:
            return constant_string_strat
        case mm.Builtins.Bool:
            return constant_bool_strat
        case _:
            raise Exception(f"No generation strategy provided for constants of type {type.name} ({type.id})")



#-------------------------------------------------------------------------------
# Scope
#-------------------------------------------------------------------------------

@dataclass
class GenScopeBase(ABC):
    """
    GenScopeBase captures only the essential data used by a GenScope.
    It's split out to allow the builder and the scope itself to maintain the same interface.
    All non-essential data will be generated when the actual GenScope is constructed by calling builder.finish()
    """
    context: GenContext = field()
    vars: list[mm.Var] = field(default_factory=list)
    entities: list[mm.Var] = field(default_factory=list)
    by_type: dict[mm.Type, list[mm.Var]] = field(default_factory=dict)

    def __str__(self):
        console = rich.get_console()
        with console.capture() as capture:
            console.rule("Scope")
            console.print("    Vars")
            for var in self.vars:
                console.print(f"    - {var.name}: {var.type.name} ({var.id})")

            console.print("    Entities")
            for var in self.entities:
                console.print(f"    - {var.name}: {var.type.name} ({var.id})")
            console.print("    By Type")
            for type, vars in self.by_type.items():
                console.print(f"    - {type.name} ({type.id})")
                for var in vars:
                    console.print(f"        - {var.name}: {var.type.name} ({var.id})")
        return capture.get()


@dataclass
class GenScope(GenScopeBase):
    """
    A GenScope stores information which is largely static over the duration of a task.
    """
    context: GenContext = field()
    vars: list[mm.Var] = field(default_factory=list)
    entities: list[mm.Var] = field(default_factory=list)
    by_type: dict[mm.Type, list[mm.Var]] = field(default_factory=dict)

    task_strategies: dict[mm.Task, Strategy] = field(init=False, default_factory=dict)
    _gen_filter: Strategy[ir.Call] = field(init=False, kw_only=True)
    _gen_computed: dict[mm.Type, Strategy[ir.Call]] = field(init=False, kw_only=True)
    _gen_agg: dict[mm.Type, Strategy[ir.Call]] = field(init=False, kw_only=True)
    _gen_arbitrary_computed: Strategy[ir.Call] = field(init=False, kw_only=True)
    _gen_arbitrary_agg: Strategy[ir.Call] = field(init=False, kw_only=True)

    def __post_init__(self):
        filter_strats: list[Strategy[ir.Call]] = []
        agg_strats: dict[mm.Type, list[Strategy[ir.Call]]] = {}
        computed_strats: dict[mm.Type, list[Strategy[ir.Call]]] = {}


        for task in self.context.tasks:
            is_agg = ir.implements(task, mm.Builtins.Aggregate)
            group = agg_strats if is_agg else computed_strats
            is_infix_filter = ir.implements(task, mm.Builtins.Infix) and task.properties[-1].is_input

            if is_agg:
                strat = self.agg_task_to_strategy(task)
            elif is_infix_filter:
                strat = self.infix_filter_task_to_strategy(task)
            else:
                strat = self.task_to_strategy(task)

            self.task_strategies[task] = strat

            is_filter = True
            for prop in task.properties:
                if not prop.is_input:
                    is_filter = False
                    group.setdefault(prop.type, []).append(strat)

            if is_filter:
                filter_strats.append(strat)

        self._gen_filter = group_limiter(gen.one_of(*filter_strats), expr=4)
        self._gen_agg = {type: gen.one_of(*strats) for type, strats in agg_strats.items() if strats}
        self._gen_computed = {type: gen.one_of(*strats) for type, strats in computed_strats.items() if strats}

        self._gen_arbitrary_agg = gen.one_of(*self._gen_agg.values())
        self._gen_arbitrary_computed = gen.one_of(*self._gen_computed.values())

    def gen_filter(self):
        return self._gen_filter

    def gen_computed(self, type: mm.Type|None):
        return self._gen_arbitrary_computed if type is None or type == mm.Builtins.Any else self._gen_computed.get(type)
        # return gen.deferred(lambda: self._gen_arbitrary_computed if type == None or type == mm.Builtins.Any else self._gen_computed[type])

    def gen_agg(self, type: mm.Type|None):
        return self._gen_arbitrary_agg if type is None or type == mm.Builtins.Any else self._gen_agg.get(type)
        # return gen.deferred(lambda: self._gen_arbitrary_agg if type == None or type == mm.Builtins.Any else self._gen_agg[type])

    def build(self):
        """Construct a builder that copies all the state from the current scope."""
        return GenScopeBuilder(
            context=self.context,
            vars=self.vars[:],
            entities=self.entities[:],
            by_type={k: v[:] for k, v in self.by_type.items()})

    def list_vars(self, type: mm.Type|None):
        """Returns a list of all vars in scope of the given type, or all of them if type is None."""
        return self.vars if type is None or type == mm.Builtins.Any else self.by_type.get(type, [])

    def try_sample_var(self, type: mm.Type|None):
        """Sample a variable of the given type, or any variable in scope if type is None."""
        vars = self.list_vars(type)
        if vars:
            return gen.sampled_from(vars)

    def sample_var(self, type: mm.Type|None):
        """Sample a variable of the given type, or any variable in scope if type is None."""
        vars = self.list_vars(type)
        if not vars:
            raise ValueError(f"Cannot sample vars of type {type}. No variables of given type in scope.")

        return gen.sampled_from(vars)

    def sample_entity(self):
        """Sample an entity variable."""
        vars = self.entities
        if not vars:
            raise ValueError("Cannot sample entity vars. No variables of given type in scope.")

        return gen.sampled_from(vars)

    def get_promotable_vars(self) -> Sequence[mm.Var]:
        """List variables which can be promoted to but are not yet entities."""
        return [v for v in self.vars if v not in self.entities and v.type not in ir.PRIMITIVE_TYPES]

    def sample_promotable_var(self, allow_empty = False):
        """Sample a variable which can be promoted to an entity but is not currently used as one."""
        vars = self.get_promotable_vars()
        if allow_empty and not vars:
            return gen.nothing()

        return gen.sampled_from(self.vars)

    def gen_value(self, type: mm.Type|None):
        strats = []
        if type in SUPPORTS_CONST_GEN or type is None:
            strats.append(gen_constant(type))
        var_strat = self.try_sample_var(type)
        if var_strat is not None:
            strats.append(var_strat)
        return gen.one_of(*strats)

        # can_be_var = self.vars and (type in self.by_type or type == None or type == mm.Builtins.Any)
        # can_be_const = type in SUPPORTS_CONST_GEN or type == None
        # if can_be_var and can_be_const:
        #     return gen.one_of(
        #         gen_constant(type),
        #         self.sample_var(type))
        # elif can_be_var:
        #     return self.sample_var(type)
        # elif can_be_const:
        #     return gen_constant(type)
        # elif type:
        #     # raise Exception(f"No vars provided for requested type {type.name} ({type.id}) which has no constant generator.")
        #     return gen.nothing()
        # else:
        #     raise Exception(f"This should never happen, but apparently the scope has no vars and we're unable to generate any constants.")

    def gen_dyn_expr(self, type: mm.Type|None):
        value_strat = self.sample_var(type)
        agg_strats = self.gen_agg(type)
        computed_strats = self.gen_computed(type)
        if agg_strats is not None and computed_strats is not None:
            return gen.one_of(value_strat, agg_strats, computed_strats)
        elif agg_strats is not None:
            return gen.one_of(value_strat, agg_strats)
        elif computed_strats is not None:
            return gen.one_of(value_strat, computed_strats)
        else:
            return value_strat

    def gen_expr_leaf(self, type: mm.Type|None):
        value_strat = self.gen_value(type)
        agg_strats = self.gen_agg(type)
        computed_strats = self.gen_computed(type)
        if agg_strats is not None and computed_strats is not None:
            return gen.one_of(value_strat, agg_strats, computed_strats)
        elif agg_strats is not None:
            return gen.one_of(value_strat, agg_strats)
        elif computed_strats is not None:
            return gen.one_of(value_strat, computed_strats)
        else:
            return value_strat

    def gen_expr(self, type: mm.Type|None, max_leaves = 6):
        return group_limiter(self.gen_expr_leaf(type), expr=max_leaves)

    def _gen_output(self, property: mm.Property):
        return gen.builds(mm.Var, type=gen.just(property.type))

    def task_to_strategy(self, task: mm.Task) -> Strategy:
        def build_args():
            first = task.properties[0]
            rest = task.properties[1:]
            try:
                return gen.tuples(self.gen_dyn_expr(first.type), *(self.gen_expr(prop.type) if prop.is_input else self._gen_output(prop) for prop in rest))
            except ValueError:
                return gen.nothing()

        def build_exhausted_args():
            first = task.properties[0]
            rest = task.properties[1:]
            try:
                return gen.tuples(self.sample_var(first.type), *(self.gen_value(prop.type) if prop.is_input else self._gen_output(prop) for prop in rest))
            except ValueError:
                return gen.nothing()

        return limited_by_group(
            "expr",
            gen.builds(ir.Call, gen.just(task), gen.deferred(build_args)),
            gen.builds(ir.Call, gen.just(task), gen.deferred(build_exhausted_args)))

    def infix_filter_task_to_strategy(self, task: mm.Task) -> Strategy:
        def build_args():
            return gen.tuples(self.sample_var(None), self.gen_expr(None))

        def build_exhausted_args():
            return gen.tuples(self.sample_var(None), self.gen_value(None))

        return limited_by_group(
            "expr",
            gen.builds(ir.Call, gen.just(task), gen.deferred(build_args)),
            gen.builds(ir.Call, gen.just(task), gen.deferred(build_exhausted_args)))

    def agg_task_to_strategy(self, task: mm.Task) -> Strategy:
        for prop in task.properties:
            if prop.name == "result":
                result_prop = prop
                break
        else:
            raise Exception(f"Could not find result property in aggregate {task.name}({[p.name for p in task.properties]}).")

        def build_args():
            if not self.vars:
                return gen.nothing()

            # @NOTE: It's possible to inline expressions here but that seems like an insane thing to do (?)
            # group = gen.lists(self.gen_expr(None), max_size=6, unique=True)
            # projection = gen.lists(self.gen_expr(None), min_size=1, max_size=6, unique=True)
            projection = gen.lists(self.sample_var(None), min_size=1, max_size=6, unique=True)
            group = gen.lists(self.sample_var(None), max_size=6, unique=True)
            result = self._gen_output(result_prop)
            return gen.tuples(projection, group, result)

        return limited_by_group(
            "expr",
            gen.builds(ir.Call, gen.just(task), gen.deferred(build_args)),
            self.gen_value(result_prop.type)) # @FIXME: This feels nasty but I don't have a better solution atm

    def gen_assert_type(self):
        return gen.builds(
            ir.AssertType,
            op = gen.sampled_from(ir.AssertOp),
            entity = self.sample_entity(),
            type = self.context.sample_non_primitive_type(),
            generated = gen.booleans())

    def gen_new_prop(self, type: mm.Type|None):
        return gen.builds(mm.Property, name=gen_python_ident(), type=gen.just(type) if type else self.context.sample_type())

    def gen_prop_for_entity(self, entity: mm.Var, only_existing = False):
        strategy = gen.sampled_from(entity.type.properties)
        if not only_existing:
            strategy = gen.one_of(strategy, self.gen_new_prop(None))

        return gen.tuples(gen.just(entity), strategy)

    def gen_ent_prop_pair(self, only_existing = False):
        return self.sample_entity().flatmap(lambda entity: self.gen_prop_for_entity(entity, only_existing))

    def gen_assert_property(self):
        return self.gen_ent_prop_pair().flatmap(lambda pair: gen.builds(
            ir.AssertProperty,
            op = gen.sampled_from(ir.AssertOp),
            entity = gen.just(pair[0]),
            property = gen.just(pair[1]),
            value = self.gen_expr(pair[1].type)))

    def gen_assert(self):
        return gen.one_of(self.gen_assert_type(), self.gen_assert_property())

@dataclass
class GenScopeBuilder(GenScopeBase):
    """
    Convenience class for constructing a GenScope. Remember that any scopes used for generation _must_ not be mutated.
    """

    finished = False
    items: list[ir.Provider] = field(default_factory=list)

    def assert_not_finished(self):
        if self.finished:
            raise BuilderConsumedError(self)

    def provide(self, item: ir.Provider):
        """Expose a new set of variables into the scope."""
        self.assert_not_finished()

        self.items.append(item)
        match item:
            case ir.EntityProvider():
                self.entities.append(item.entity)
                self.vars += item.provides
                for v in item.provides:
                    self.by_type.setdefault(v.type, []).append(v)
            case ir.ComputedProvider():
                self.vars += item.provides
                for v in item.provides:
                    self.by_type.setdefault(v.type, []).append(v)
            case _:
                raise Exception(f"GenScopeBuilder does not know how to handle provider of type {type(item)}")

    def finish(self) -> GenScope:
        """Efficiently drain the builder's state into a scope. Note that it _cannot_ be reused after this."""
        self.finished = True
        return GenScope(
            context=self.context,
            vars=self.vars,
            entities=self.entities,
            by_type = self.by_type)

    def snapshot(self) -> GenScope:
        """Construct a scope from the current builder state without consuming the builder."""
        return GenScope(
            context=self.context,
            vars=self.vars[:],
            entities=self.entities[:],
            by_type = {k: v[:] for k, v in self.by_type.items()})



import relationalai as rai 
from relationalai import std


model = rai.Model(name=globals().get("name", ""), config=globals().get("config"))
Object = model.Type("Object")
Relationship = model.Type("Relationship")

with model.rule():
    obj1 = Object.add(id=1)
    obj2 = Object.add(id=2)
    Relationship.add(from_=obj1, to=obj2)

disconnected_graph = std.graphs.Graph(model)
disconnected_graph.Node.extend(Object)

connected_graph = std.graphs.Graph(model)
with model.rule():
    r = Relationship()
    connected_graph.Edge.add(r.from_, r.to)

# Check that is_connected works
with model.query() as select:
    with model.match() as connected:
        with connected_graph.compute.is_connected():
            connected.add(True)
        with model.case():
            connected.add(False)
    with model.match() as disconnected:
        with disconnected_graph.compute.is_connected():
            disconnected.add(True)
        with model.case():
            disconnected.add(False)
    select(connected, disconnected)

# Check that reachable_from works
with model.query() as select:
    n = Object()
    reachable_from_n = connected_graph.compute.reachable_from(n)
    select(n, reachable_from_n)

# Check that is_reachable works
with model.query() as select:
    n1 = Object(id=1)
    n2 = Object(id=2)
    with model.match() as reachable1:
        with model.case():
            connected_graph.compute.is_reachable(n1, n2)
            reachable1.add(True)
        with model.case():
            reachable1.add(False)
    with model.match() as reachable2:
        with model.case():
            connected_graph.compute.is_reachable(n2, n1)
            reachable2.add(True)
        with model.case():
            reachable2.add(False)
    select(reachable1, reachable2)



import urllib
import re

from relationalai.errors import RAIException

# replace the values of the URL parameters that start with X-Amz- with XXX
def scrub_url(url):
    parsed = urllib.parse.urlparse(url)
    parsed_qs = urllib.parse.parse_qs(parsed.query)
    for key in parsed_qs:
        if key.startswith("X-Amz-"):
            parsed_qs[key] = "XXX"
    new_qs = urllib.parse.urlencode(parsed_qs, doseq=True)
    return urllib.parse.urlunparse(parsed._replace(query=new_qs))

def find_urls(string):
    url_pattern = re.compile(r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')
    urls = re.findall(url_pattern, string)
    return urls

def scrub_urls(string, urls):
    for url in urls:
        # replace with scrubbed version
        string = string.replace(url, scrub_url(url))
    return string

def scrub_exception(exception):
    exception_str = str(exception)
    urls = find_urls(exception_str)
    if urls:
        return RAIException(scrub_urls(exception_str, urls))
    return exception

def escape_for_f_string(code: str) -> str:
    return (
        code
        .replace("\\", "\\\\")
        .replace("{", "{{")
        .replace("}", "}}")
        .replace("\n", "\\n")
        .replace('"', '\\"')
        .replace("'", "\\'")
    )

def escape_for_sproc(code: str) -> str:
    return code.replace("$$", "\$\$")


# pyright: reportUnusedExpression=false
import relationalai as rai

model = rai.Model(name=globals().get("name", "test_smoke"), config=globals().get("config"))
Person = model.Type("Person")
Adult = model.Type("Adult")

with model.rule():
    Person.add(name="Joe", age=74)
    Person.add(name="Bob", age=40)
    Person.add(name="Jane", age=10)

with model.rule():
    p = Person()
    p.age >= 18
    p.set(Adult)

with model.rule():
    p = Person()
    with p.age == 10:
        p.set(coolness=100)

with model.query() as select:
    a = Adult()
    10 <= a.age <= 80
    z = select(a, a.name, a.age)
print(z.results.dtypes)

with model.query() as select:
    p = Person()
    with model.found():
        p.coolness
    z = select(p, p.coolness)

print(z.results.dtypes)



from abc import ABC, abstractmethod
from typing import Callable, Generic, TypeVar
from colorama import Back, Fore, Style
from gentest.harness.database import get_short_path, get_value_hash
from gentest.util import condense_traceback, decorate, fmt_h3, fmt_indented, fmt_lines, fmt_styled_lines, get_last_exception_entry
from hypothesis.errors import _Trimmable

T = TypeVar("T")

Formatter = Callable[[T], str]

ERROR_DECORATION = decorate(Back.RED, Back.RESET, Fore.RED + Style.BRIGHT, Style.NORMAL + Fore.RESET)

#===============================================================================
# Errors
#===============================================================================

class LazyException(Exception, ABC):
    def __init__(self, short: str):
        super().__init__(short)
        self.short = short
        self.long = None
        self.message = short

    @abstractmethod
    def pprint(self) -> str:
        pass

    def reify(self):
        if not self.long:
            self.long = self.pprint()
            self.message = self.long

        return self

    def __str__(self):
        return self.message

class GenException(LazyException):
    conjecture: bytes

    label: str|None
    key_hash: str|None

    def __init__(self, short: str, conjecture: bytes, key_hash: str|None = None):
        super().__init__(short)
        self.conjecture = conjecture
        self.key_hash = key_hash

    def annotate(self, label: str|None, key_hash: str|None):
        self.label = label
        self.key_hash = key_hash

    def value_hash(self):
        return get_value_hash(self.conjecture)

    def hash_path(self, key_hash: str|None = None):
        if not key_hash:
            key_hash = self.key_hash

        if not key_hash:
            print("Cannot get hash path without a provided or annotated key hash.")
            return ""

        return get_short_path(key_hash, self.value_hash())

class RoundtripError(GenException, Generic[T], ABC):
    description: str
    expected: T

    def __init__(self, expected: T, conjecture: bytes):
        name = self.__class__.__name__
        short = f"{name}: {self.description}"
        super().__init__(short, conjecture)
        self.expected = expected

    @abstractmethod
    def body(self) -> list[str]:
        pass

    def pprint(self):
        return "\n".join([
            f"{Style.BRIGHT}{self.short}{Style.RESET_ALL}",
            *self.body()
        ])

class WrappedError(RoundtripError[T]):
    description = "An unexpected error occurred while running the test."

    def __init__(self, expected: T, conjecture: bytes, err: Exception, fmt: Formatter[T] = str):
        super().__init__(expected, conjecture)
        self.err = err
        self.fmt = fmt

    def body(self):
        pad = "    "
        return [
            fmt_h3("CAUSE", indent=1),
            fmt_indented(str(self.err) + ("\n" + condense_traceback(self.err)), 3),
            fmt_h3("MODEL", indent=1),
            fmt_lines(self.fmt(self.expected), pad)]

class EmitError(RoundtripError[T]):
    description = "Failed to emit code for model."

    def __init__(self, expected: T, conjecture: bytes, err: Exception, fmt: Formatter[T] = str):
        self.description = f"Failed to emit code for {expected.__class__.__name__}."
        super().__init__(expected, conjecture)
        self.err = err
        self.fmt = fmt

    def body(self):
        pad = "    "
        return [
            fmt_h3("CAUSE", indent=1),
            fmt_indented(str(self.err) + ("\n" + condense_traceback(self.err)), 4),
            fmt_h3("MODEL", indent=1),
            fmt_lines(self.fmt(self.expected), pad)]

class ExecError(RoundtripError[T]):
    description = "Failed to execute emitted PyRel."

    def __init__(self, expected: T, conjecture: bytes, code: str, err: Exception, fmt: Formatter[T] = str):
        super().__init__(expected, conjecture)

        self.code = code
        self.err = err
        self.fmt = fmt

        filename, err_line_ix, _ = get_last_exception_entry(err)
        if filename == "<string>":
            self.err_line_ix = err_line_ix
        else:
            self.err_line_ix = None

    def body(self):
        pad = "    "
        return [
            fmt_h3("CAUSE", indent=1),
            fmt_indented(str(self.err) + ("\n" + condense_traceback(self.err) if self.err_line_ix is None else ""), 4),
            fmt_h3("MODEL", indent=1),
            fmt_lines(self.fmt(self.expected), pad),
            fmt_h3("PYREL", indent=1),
            fmt_styled_lines(self.code, {self.err_line_ix: ERROR_DECORATION}, pad)]

class DiffError(RoundtripError[T]):
    def __init__(self, expected: T, conjecture: bytes, code: str, actual: T, diff: str):
        self.description = f"{expected.__class__.__name__} diverged from original."
        super().__init__(expected, conjecture)
        self.code = code
        self.actual = actual
        self.diff = diff

    def body(self):
        pad = "  "
        return [
            fmt_h3("DELTA", indent=0),
            fmt_lines(self.diff, pad),
            fmt_h3("PYREL", indent=0),
            fmt_lines(self.code, pad)]


def reify(err: Exception, key_hash: str) -> Exception:
    match err:
        case ExceptionGroup():
            for child in err.exceptions:
                reify(child, key_hash)
            return err
        case GenException():
            err.annotate(None, key_hash)
            return err.reify()
        case _Trimmable():
            err.__traceback__ = None
            return err
        case _:
            return err



import utils

import relationalai as rai
from relationalai.std.graphs import Graph

# --------------------------------------------------
# Load data
# --------------------------------------------------
articles = utils.get_articles(num_articles=5, use_local=True)

# --------------------------------------------------
# Types
# --------------------------------------------------
model = rai.Model("ArticleAnalysis")
Article = model.Type("Article")
ContentSource = model.Type("ContentSource")
Language = model.Type("Language")
Author = model.Type("Author")

# --------------------------------------------------
# Load data
# --------------------------------------------------
with model.rule(dynamic=True):
    for article in articles:
        source = ContentSource.add(name=article["source"]["title"])
        language = Language.add(name=article["lang"])
        author = Author.add(name=utils.get_author(article))

        a = Article.add(
            title=article.get("title", "Unknown Title"),
            trimmed_title=utils.trim_node_label(article.get("title", "Unknown Title")),
            sentiment=article["sentiment"] or "N/A",
            body=article["body"],
            dType=article["dataType"],
            published_on=article["dateTimePub"],
            date=article["date"],
            url=article["url"],
            language=language,
            author=author,
            source=source,
        )

# --------------------------------------------------
# Query for articles to see what we have
# --------------------------------------------------
with model.query() as select:
    a = Article()
    response = select(a.title, a.body, a.date, a.source)
print(response.results)

# --------------------------------------------------
# Graph
# --------------------------------------------------
graph = Graph(model)
Node, Edge = graph.Node, graph.Edge

with model.rule():
    a = Article()
    Node.add(a, label=a.trimmed_title, type="article")
    Node.add(a.source, label=a.source.name, type="source")
    Node.add(a.language, label=a.language.name, type="language")
    Node.add(a.author, label=a.author.name, type="author")
    Edge.add(a, a.source, type="publishedBy")
    Edge.add(a, a.language, type="writtenIn")
    Edge.add(a, a.author, type="writtenBy")

graph.visualize(
    three=False,
    show_edge_label=True,
    node_label_size_factor=1.5,
    node_size_factor=1.5,
    node_hover_neighborhood=True,
    show_node_label=True,
    style={
        "node": {
            "size": utils.get_node_size_by_type,
            "color": utils.get_node_color_by_type,
        },
        "edge": {
            "opacity": 0.8,
            "color": utils.get_edge_color_by_type,
            "label": lambda t: f'{t["type"]}',
        },
    },
).display()



from _pytest._io.wcwidth import wcswidth
from _pytest.terminal import TerminalReporter
from gentest.util import condense_traceback
import pytest
import pandas as pd
pd.set_option('display.max_columns', None)
pd.set_option('display.expand_frame_repr', False)

@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    term: TerminalReporter = item.config.pluginmanager.getplugin("terminalreporter")
    write_status(term)

    yield


def write_status(term: TerminalReporter):
    cur_line = term._tw._current_line
    prefix = trim_suffix(cur_line)
    w = wcswidth(prefix) # term._width_of_current_line
    term._tw.fullwidth - w - 1

    term.rewrite("\r" + prefix, erase=True)
    # term.write(msg.rjust(fill), flush=True, green=True)

def trim_suffix(line: str):
    ix = line.find("   ")
    return line if ix == -1 else line[0:ix]



def custom_repr_failure(self, excinfo: pytest.ExceptionInfo, style=None):
    err = excinfo.value
    match err:
        case Exception():
            return str(err) + "\n\n" + condense_traceback(err)
        case _:
            raise err
    # return str(excinfo.value)
    # print("hi", excinfo)

pytest.Function.repr_failure = custom_repr_failure



import os
import shutil
import subprocess

simple_program_body = """
Person = model.Type("Person")

with model.rule():
    alex = Person.add(name="Alex", age=19)
    bob = Person.add(name="Bob", age=47)
    carol = Person.add(name="Carol", age=17)
    deb = Person.add(name="Deb", age=17)

    carol.set(friend=deb)
    alex.set(friend=bob)
    alex.set(friend=carol)

with model.query() as select:
    alex = Person(name="Alex")
    person = alex.friend
    person.age >= 21
    response = select(person.name)

print(response.results)
# Output:
#   name
# 0  Bob
"""

hello_world_pyrel_example = f"""
import relationalai as rai

model = rai.Model("people")

{simple_program_body}
"""

hello_world_expected_output = """
  name
0  Bob
"""

example_with_profile = f"""
import relationalai as rai

model = rai.Model("people", profile="test")

{simple_program_body}
"""

example_with_explicit_config = f"""
import relationalai as rai
from relationalai.clients import config as cfg

config = cfg.Config({{
    "platform": "azure",
    "host": "azure.relationalai.com",
    "port": 443,
    "region": "us-east",
    "scheme": "https",
    "client_credentials_url": "https://login.relationalai.com/oauth/token",
    "client_id": "{os.getenv("RAI_CLIENT_ID")}",
    "client_secret": "{os.getenv("RAI_CLIENT_SECRET")}",
    "engine": "aaa",
}})

model = rai.Model("people", config=config)

{simple_program_body}
"""

def get_project_dir():
    dir = os.path.abspath(__file__)
    for i in range(3):
        dir = os.path.dirname(dir)
    return dir

def setup_tempdir(replace:bool=True):
    tempdir = os.path.expanduser("~/temp-relationalai-test")
    if replace:
      if os.path.exists(tempdir) and os.path.isdir(tempdir):
          shutil.rmtree(tempdir)
      os.mkdir(tempdir)
      os.chdir(tempdir)
    elif not os.path.exists(tempdir):
      os.mkdir(tempdir)
      os.chdir(tempdir)
    if not replace:
       return tempdir

    print("Current working directory:", os.getcwd())
    print("Project directory:", get_project_dir())
    print("Creating a virtual environment...")
    subprocess.run(["python", "-m", "venv", ".venv"], capture_output=True)
    print("Installing relationalai-python...")
    subprocess.run(
        [".venv/bin/python", "-m", "pip", "install", "-e", get_project_dir()],
        capture_output=True,
    )
    return tempdir


from typing import Sequence
from hypothesis import given, settings, strategies as gen
from gentest.emit import DocumentEmitter
from relationalai import metamodel as mm
from relationalai.clients.test import Document, Install, Query
from gentest.gen.scope import GenScope
from gentest.gen.task import gen_task
from gentest import fixtures
from gentest.gen.context import fixture

def gen_query(scope: GenScope):
    return gen_task(scope).map(lambda task: ("query", task)) # @FIXME

def gen_rule(scope: GenScope):
    return gen_task(scope).map(lambda task: ("install", task))

def gen_block(scope: GenScope):
    return gen_rule(scope) | gen_query(scope)

def gen_document(root_scope: GenScope):
    return gen.lists(gen_block(root_scope), min_size=1, max_size=10).map(into_document)


def into_document(blocks: Sequence[tuple[str, mm.Task]]) -> Document:
    doc = Document()
    for (type, task) in blocks:
        match type:
            case "query":
                doc.blocks.append(Query(task, doc.query_count))
                doc.query_count += 1
            case "install":
                doc.blocks.append(Install(task))
            case _:
                raise Exception(f"Unknown block type {type} for task {str(task)}")

    return doc

if __name__ == "__main__":
    ctx = fixture(fixtures.person_place_thing).finish()
    root = GenScope(ctx)

    @settings(max_examples=200)
    @given(gen_document(root))
    def foo(doc: Document):
        print("="*80)
        print("blocks:", len(doc.blocks))
        print(DocumentEmitter.from_document(doc).stringify())
        print("="*80)
    foo()



import relationalai as rai
from relationalai.std import alias

model = rai.Model('EVPenetrationExample')
State = model.Type('State')

# Load data from azure blob storage before running the script
# rai imports:snapshot --source "azure://raidocs.blob.core.windows.net/datasets/states/state_stats.csv" schema:state_name=string schema:area=int schema:population=int schema:ev_registration_count=int

# Or load data from a local CSV file using either a relative path, as below, or an absolute path.
# rai imports:snapshot --source "examples/ev_penetration/state_stats.csv" schema:state_name=string schema:area=int schema:population=int schema:ev_registration_count=int

# Read data from csv file
with model.read("state_stats.csv") as row:
    state = State.add(name=row.state_name).set(area=row.area, population=row.population, ev_registration=row.ev_registration_count)

# Query the collection
with model.query() as select:
    state = State()
    ev_penetration = 1000 * state.ev_registration / state.population
    response = select(
        state.name,
        state.population,
        state.ev_registration,
        alias(ev_penetration, "ev_penetration"),
    )

# Display the results sorted in descending order by the `ev_penetration` column
print(response.results.sort_values(by="ev_penetration", ascending=False))




import relationalai as rai
from relationalai import std

# Query snapshots are output in same order as this list.
UNWEIGHTED_ALGOS = [
    "jaccard_similarity",
    "cosine_similarity",
]

WEIGHTED_ALGOS = [
    "weighted_jaccard_similarity",
    "weighted_cosine_similarity",
]

model = rai.Model(name=globals().get("name", ""), config=globals().get("config"))
Object = model.Type("Object")
Relationship = model.Type("Relationship")

with model.rule():
    obj1 = Object.add(id=1)
    obj2 = Object.add(id=2)
    Relationship.add(from_=obj1, to=obj2, weight=1.0)

# Make unweighted and weighted graphs.
# NOTE: cosine_similarity only supports undirected graphs.
unweighted_graph = std.graphs.Graph(model, undirected=True)
with model.rule():
    r = Relationship()
    unweighted_graph.Edge.add(r.from_, r.to)

weighted_graph = std.graphs.Graph(model, undirected=True, weighted=True)
with model.rule():
    r = Relationship()
    weighted_graph.Edge.add(r.from_, r.to, weight=r.weight)

# Test the unweighted algorithms.
for algo_name in UNWEIGHTED_ALGOS:
    with model.query(tag=algo_name) as select:
        o1 = Object()
        o2 = Object()
        algo = getattr(unweighted_graph.compute, algo_name)(o1, o2)
        select(o1, o2, algo)

# Test the weighted algorithms.
for algo_name in WEIGHTED_ALGOS:
    with model.query(tag=algo_name) as select:
        o1 = Object()
        o2 = Object()
        algo = getattr(weighted_graph.compute, algo_name)(o1, o2)
        select(o1, o2, algo)



import pytest
import relationalai as rai
from relationalai.std.graphs import Graph

DEFAULT_METADATA = {
    'arrow_color': '#999',
    'arrow_size': 4,
    'edge_border_color': '#999',
    'edge_border_size': 1,
    'edge_color': '#999',
    'edge_label_color': 'black',
    'edge_label_size': 10,
    'edge_opacity': 1,
    'edge_shape': 'circle',
    'edge_size': 2,
    'node_border_color': 'black',
    'node_border_size': 1,
    'node_color': 'black',
    'node_label_color': 'black',
    'node_label_size': 10,
    'node_opacity': 1,
    'node_shape': 'circle',
    'node_size': 10,
}


def make_graph(model, test_input: dict):
    Node = model.Type("Node")

     # Add objects to the model.
    with model.rule(dynamic=True):
        for (id1, adj) in test_input["adj_list"].items():
            node1 = Node.add(id=id1)
            for id2 in adj:
                node2 = Node.add(id=id2)
                node1.set(adjacent_to=node2)

    # Create a graph object.
    graph = Graph(model, undirected=test_input["undirected"])

    # Add nodes to the graph.
    with model.rule(dynamic=True):
        for (id, props) in test_input["node_props"].items():
            node = Node(id=id)
            graph.Node.add(node, **props)

    # Add edges to the graph.
    with model.rule(dynamic=True):
        for ((id1, id2), props) in test_input["edge_props"].items():
            node1, node2 = Node(id=id1), Node(id=id2)
            graph.Edge.add(node1, node2, **props)

    return graph

# Test that graph._visual_dict() returns a dictionary with the correct keys and values
@pytest.mark.parametrize(
    ["test_input", "expected_dict"],
    [
        # Undirected graph with one data edge, no props, default style
        pytest.param(
            {
                "adj_list": {1: [2]},
                "node_props": {1: {}, 2: {}},
                "edge_props": {(1, 2): {}},
                "undirected": True,
                "style": {},
                "kwargs": {},
            },
            {
                'directed': False,
                'nodes': {
                    'jLWfdS5w49/Nqwx+mYthYw': {'metadata': {}},
                    'peH/vEX2TeRy3NL41/2Nag': {'metadata': {}},
                },
                'edges': [{'metadata': {}, 'source': 'jLWfdS5w49/Nqwx+mYthYw', 'target': 'peH/vEX2TeRy3NL41/2Nag'}],
                'metadata': DEFAULT_METADATA,
            },
            id="undirected_no_props_default_style",
        ),
        # Directed graph with one data edge, no props, no style
        pytest.param(
            {
                "adj_list": {1: [2]},
                "node_props": {1: {}, 2: {}},
                "edge_props": {(1, 2): {}},
                "undirected": False,
                "style": {},
                "kwargs": {},
            },
            {
                'directed': True,
                'nodes': {
                    'jLWfdS5w49/Nqwx+mYthYw': {'metadata': {}},
                    'peH/vEX2TeRy3NL41/2Nag': {'metadata': {}},
                },
                'edges': [
                    {'metadata': {}, 'source': 'peH/vEX2TeRy3NL41/2Nag', 'target': 'jLWfdS5w49/Nqwx+mYthYw'},
                ],
                'metadata': DEFAULT_METADATA,
            },
            id="directed_no_props_default_style",
        ),
        # Undirected graph with two data edges (but one graph edge), props, custom node and edge colors.
        # Shows property merge behavior for undirected graphs with multiple data edges for a single graph edge.
        # Compare to the directed graph case below.
        pytest.param(
            {
                "adj_list": {1: [2]},
                "node_props": {1: {'node_color': 'blue'}, 2: {'node_color': 'red'}},
                "edge_props": {
                    (1, 2): {'color': 'green'},
                    (2, 1): {'color': 'yellow'}
                },
                "undirected": True,
                "style": {
                    'nodes': {'color': lambda n: n['node_color']},
                    'edges': {'color': lambda e: e['color']}
                },
                "kwargs": {},
            },
            {
                'directed': False,
                'nodes': {
                    'jLWfdS5w49/Nqwx+mYthYw': {'metadata': {'node_color': 'red'}},
                    'peH/vEX2TeRy3NL41/2Nag': {'metadata': {'node_color': 'blue'}},
                },
                'edges': [{'metadata': {'color': 'green'}, 'source': 'jLWfdS5w49/Nqwx+mYthYw', 'target': 'peH/vEX2TeRy3NL41/2Nag'}],
                'metadata': DEFAULT_METADATA,
            },
            id="undirected_with_props_custom_style",
        ),
        # Directed graph with two data edges (and two graph edges), props, custom node and edge colors.
        pytest.param(
            {
                "adj_list": {1: [2]},
                "node_props": {1: {'node_color': 'blue'}, 2: {'node_color': 'red'}},
                "edge_props": {
                    (1, 2): {'color': 'green'},
                    (2, 1): {'color': 'yellow'}
                },
                "undirected": False,
                "style": {
                    'nodes': {'color': lambda n: n['node_color']},
                    'edges': {'color': lambda e: e['color']}
                },
                "kwargs": {},
            },
            {
                'directed': True,
                'nodes': {
                    'jLWfdS5w49/Nqwx+mYthYw': {'metadata': {'node_color': 'red'}},
                    'peH/vEX2TeRy3NL41/2Nag': {'metadata': {'node_color': 'blue'}},
                },
                'edges': [
                    {'metadata': {'color': 'yellow'}, 'source': 'jLWfdS5w49/Nqwx+mYthYw', 'target': 'peH/vEX2TeRy3NL41/2Nag'},
                    {'metadata': {'color': 'green'}, 'source': 'peH/vEX2TeRy3NL41/2Nag', 'target': 'jLWfdS5w49/Nqwx+mYthYw'}
                ],
                'metadata': DEFAULT_METADATA,
            },
            id="directed_with_props_custom_style",
        ),
    ],
)
def test_visual_dict(engine_config, test_input, expected_dict):
    model_name = "testGraphVisualDict"
    model = rai.Model(model_name, config=engine_config)
    graph = make_graph(model, test_input)
    visual_dict = graph._visual_dict(style=test_input["style"])
    assert visual_dict == {'graph': expected_dict}

# Test that graph.visualize() returns a value and does not raise an exception.
@pytest.mark.parametrize(
        ["test_input"],
        [
            pytest.param(
                {
                    "adj_list": {1: [2]},
                    "node_props": {1: {}, 2: {}},
                    "edge_props": {(1, 2): {}},
                    "undirected": True,
                    "style": {},
                    "kwargs": {},
                },
                id="no_style_no_kwargs",
            ),
            pytest.param(
                {
                    "adj_list": {1: [2]},
                    "node_props": {1: {"id": "node1"}, 2: {"id": "node2"}},
                    "edge_props": {(1, 2): {"id": "edge1"}},
                    "undirected": True,
                    "style": {},
                    "kwargs": {
                        "edge_label_data_source": "id",
                        "node_label_data_source": "id",
                        "edge_curvature": 1.0,
                    },
                },
                id="no_style_override_kwargs",
            ),
        ]
)
def test_visualize(engine_config, test_input):
    model_name = "testGraphVisualize"
    model = rai.Model(model_name, config=engine_config)
    graph = make_graph(model, test_input)
    assert graph.visualize(style=test_input["style"], **test_input["kwargs"]) is not None



#pyright: reportUnusedExpression=false
import relationalai as rai

model = rai.Model(name=globals().get("name", "test_union"), config=globals().get("config"))
Person = model.Type("Person")
Adult = model.Type("Adult")

with model.rule():
    Person.add(name="Joe", age=84)
    Person.add(name="Bob", age=40)
    Person.add(name="Jane", age=10)

#--------------------------------------------------
# Make sure shared vars between match branches stay
# in their respective branches - this would previously
# error in dsl.py
#--------------------------------------------------

with model.rule():
    p = Person()
    m = p.age
    with model.match():
        with m.name == "Per Unit":
            m.set(calculated_surcharge_rate=100)
        with m.name == "By Percentage":
            m.set(calculated_surcharge_rate=1000)

#--------------------------------------------------
# Basic union
#--------------------------------------------------

with model.query() as select:
    p = Person()
    with model.match(multiple=True) as cool:
        with p.age > 80:
            cool.add("amazing", rating=1000)
        with p.age > 60:
            cool.add("rad", rating=100)
        with p.age > 18:
            cool.add("awesome", rating=10)
    z = select(p.name, cool, cool.rating)

print(z.results)

#--------------------------------------------------
# Basic ordered choice
#--------------------------------------------------

with model.query() as select:
    p = Person()
    with model.match() as cool:
        with p.age > 80:
            cool.add("amazing", rating=1000)
        with p.age > 60:
            cool.add("rad", rating=100)
        with p.age > 18:
            cool.add("awesome", rating=10)
    z = select(p.name, cool, cool.rating)

print(z.results)

#--------------------------------------------------
# Ordered choice with effects and no returns
#--------------------------------------------------

with model.rule():
    p = Person()
    with model.match() as w:
        with model.case():
            p.age > 10
            p.set(zomg=10)
        with model.case():
            # p.age <= 10
            p.set(zomg=1)

with model.query() as select:
    p = Person()
    z = select(p, p.zomg)

print(z.results)


#--------------------------------------------------
# Constants
#--------------------------------------------------

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
PROFILE = None

SNOWFLAKE = "Snowflake"
AZURE = "Azure (Beta)"

AZURE_ENVS = {
    "Production": {
        "host": "azure.relationalai.com",
        "client_credentials_url": "https://login.relationalai.com/oauth/token"
    },
    "Early Access": {
        "host": "azure-ea.relationalai.com",
        "client_credentials_url": "https://login-ea.relationalai.com/oauth/token"
    },
    "Staging": {
        "host": "azure-staging.relationalai.com",
        "client_credentials_url": "https://login-staging.relationalai.com/oauth/token"
    },
    "Latest": {
        "host": "azure-latest.relationalai.com",
        "client_credentials_url": "https://login-latest.relationalai.com/oauth/token"
    },
}



import relationalai as rai
from relationalai import std

# Query snapshots are output in same order as this list.
ALGOS = [
    "adamic_adar",
    "preferential_attachment",
    "common_neighbor",
]

model = rai.Model(name=globals().get("name", ""), config=globals().get("config"))
Object = model.Type("Object")
LinkPrediction = model.Type("LinkPrediction")

with model.rule():
    obj1 = Object.add(id=1)
    obj2 = Object.add(id=2)
    obj3 = Object.add(id=3)
    obj4 = Object.add(id=4)
    LinkPrediction.add(from_=obj1, to=obj2)
    LinkPrediction.add(from_=obj2, to=obj3)
    LinkPrediction.add(from_=obj3, to=obj3)
    LinkPrediction.add(from_=obj2, to=obj4)
    LinkPrediction.add(from_=obj4, to=obj3)

graph = std.graphs.Graph(model, undirected=True)
Node, Edge = graph.Node, graph.Edge

Node.extend(Object)

with model.rule():
    r = LinkPrediction()
    Edge.add(r.from_, r.to)

for algo_name in ALGOS:
    with model.query(tag=algo_name) as select:
        o1 = Object(id=1)
        o2 = Object(id=2)
        o3 = Object(id=3)
        o4 = Object(id=4)
        algo1 = getattr(graph.compute, algo_name)(o1, o3)
        algo2 = getattr(graph.compute, algo_name)(o2, o4)
        select(o1, o3, algo1)
        select(o2, o4, algo2)



from collections import defaultdict
import numbers
from typing import Any, Dict, Iterable, List, Set, cast

from .rel_emitter import Emitter

from .metamodel import Behavior, Builtins, Action, ActionType, Type, Var, Task
from . import metamodel as m
from . import compiler as c
from .clients import config
from .dsl import build
from . import dsl

gather_vars = m.Utils.gather_vars
gather_task_vars = m.Utils.gather_task_vars

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def is_static(x:Any):
    if isinstance(x, m.Action):
        return all([is_static(z) for z in x.bindings.values()])
    if isinstance(x, Var):
        return x.value is not None
    if isinstance(x, Type):
        return True
    if isinstance(x, str):
        return True
    if isinstance(x, numbers.Number):
        return True
    if isinstance(x, list):
        return all(is_static(i) for i in x)
    if isinstance(x, tuple):
        return all(is_static(i) for i in x)
    if isinstance(x, dict):
        return all(is_static(i) for i in x.values())
    return False

def prepend_bindings(prefix:List[Var], action:Action):
    neue_bindings = {}
    for i, var in enumerate(prefix):
        neue_bindings[Builtins.Relation.properties[i]] = var
    prop_len = len(neue_bindings)
    for v in action.bindings.values():
        neue_bindings[Builtins.Relation.properties[prop_len]] = v
        prop_len += 1
    action.bindings = neue_bindings

#--------------------------------------------------
# OrderedSet
#--------------------------------------------------

class OrderedSet:
    def __init__(self):
        self.set:Set[Var] = set()
        self.list:List[Var] = []

    def add(self, item):
        if item not in self.set:
            self.set.add(item)
            self.list.append(item)

    def update(self, items:Iterable[Any]):
        for item in items:
            self.add(item)

    def __getitem__(self, ix):
        return self.list[ix]

    def __contains__(self, item):
        return item in self.set

    def __bool__(self):
        return bool(self.set)

    def __iter__(self):
        return iter(self.list)

#--------------------------------------------------
# FlowFrame
#--------------------------------------------------

class FlowFrame:
    multi_names = set()

    def __init__(self, task:Task, prev:'FlowFrame|None'=None):
        self.task = task
        self.finalized = False
        self.demands:Dict[Var, List[Action]] = defaultdict(list)
        self.flushed_vars = set()
        self.keys:Dict[Var, Var] = {}
        self.multi = set()
        self.prev = prev
        self.merge(prev)

    def merge(self, other:'FlowFrame|None'):
        if not other:
            return

        for (k, vs) in other.demands.items():
            for v in vs:
                if v not in self.demands[k]:
                    self.demands[k].append(v)
        for (k, v) in other.keys.items():
            self.keys[k] = v
        self.multi.update(other.multi)
        self.flushed_vars.update(other.flushed_vars)

    def push_demand(self, var:Var, action:Action, with_key=None):
        for a in self.demands[var]:
            if a.equiv(action):
                return
        self.demands[var].append(action)
        if with_key:
            self.push_key(var, with_key)

    def demand(self, var):
        ret = self.demands.get(var, [])
        if var in self.demands:
            del self.demands[var]
        self.flushed_vars.add(var)
        return ret

    def has_demand(self, var):
        return var in self.demands or var in self.flushed_vars

    def is_unique_demand(self, var, action:Action):
        for a in self.demands[var]:
            if a.equiv(action):
                return False
        return True

    def demands_available(self):
        return set(self.demands.keys())

    def check_multi_prop(self, prop:m.Property, var:Var|None=None):
        # For queries we need to know if a property is multi-valued so we can include
        # it in the key of the split up return. This doesn't apply to rules, so by the
        # time a query is called if we don't know that a prop is multi, it doesn't really
        # matter since it can't affect the results.
        if prop.name in self.multi_names or Builtins.MultiValued in prop.parents:
            self.multi_names.add(prop.name)
            if var and var.value is None:
                self.multi.add(var)
            return True
        return False

    def is_multi(self, var:Var):
        return var in self.multi

    def push_key(self, var:Var, key:Var):
        self.keys[var] = key

    def key(self, var, default):
        return self.keys.get(var, default)

    def has_action(self, action:Action):
        for a in self.task.items:
            if a.equiv(action):
                return True
        return False

    def keys_recursive(self, var):
        key = self.keys.get(var)
        if key and key != var:
            recur = self.keys_recursive(key)
            if key not in recur:
                return recur + [key]
            elif recur:
                return recur
            else:
                return [key]
        elif key:
            return [key]
        return []

    def is_empty_continue(self):
        if len(self.task.items) == 1 \
            and self.prev \
            and self.task.items[0].action == ActionType.Get \
            and self.task.items[0].entity.value == self.prev.task:
            return True

    def is_empty(self):
        return not len(self.task.items) and not len(self.task.bindings)

    def create_get(self, var_mapping:Dict[Var, Var]={}) -> Action:
        if Builtins.InlineAnnotation not in self.task.parents:
            self.task.parents.append(Builtins.InlineAnnotation)
        vars = gather_task_vars(self.task, only_direct_children=True)
        if not self.finalized:
            self.finalized = True
            self.task.items.append(build.relation_action(ActionType.Bind, self.task, vars))
        return build.relation_action(ActionType.Get, self.task, [var_mapping.get(v, v) for v in vars])

#--------------------------------------------------
# Flow
#--------------------------------------------------

class Flow():
    def __init__(self):
        self.tasks:List[Task] = []
        self.frames:List[FlowFrame] = []

    def frame(self):
        return self.frames[-1]

    def reset(self):
        self.tasks.clear()
        self.frames = []

    def _prev_frame(self, pop=False):
        if not self.frames:
            return None
        prev_frame = self.frames.pop() if pop else self.frames[-1]
        if prev_frame and prev_frame.is_empty_continue():
            grand_parent = prev_frame.prev
            if grand_parent:
                grand_parent.merge(prev_frame)
            if prev_frame.task in self.tasks:
                self.tasks.remove(prev_frame.task)
            return grand_parent
        return prev_frame

    def branch(self, inline=False, ignore_deps=False, var_mapping:Dict[Var,Var]={}):
        neue = Task()
        prev_frame = self._prev_frame()
        self.frames.append(FlowFrame(neue, prev_frame))
        if not inline:
            self.tasks.append(neue)
        if not ignore_deps and prev_frame and not prev_frame.is_empty():
            self.append(prev_frame.create_get(var_mapping))
        return neue

    def join(self, discard=False):
        frame = self.frames.pop()
        if discard:
            self.tasks.remove(frame.task)
        return frame.task

    def continue_(self):
        # remove the original root frame, and use it as the root of the continuation frame
        prev = cast(FlowFrame, self._prev_frame(pop=True))
        neue = Task()
        self.frames.append(FlowFrame(neue, prev))
        if not prev.is_empty():
            self.append(prev.create_get())
        self.tasks.append(neue)
        return neue

    def append(self, action:Action):
        task = self.frame().task
        for item in task.items:
            if item.equiv(action):
                return
        task.items.append(action)

    def available_vars(self):
        frame = self.frame()
        return gather_task_vars(frame.task) | frame.demands_available()

    def finalize(self):
        final_tasks = []
        for task in self.tasks[:-1]:
            if len(task.items) or len(task.bindings):
                final_tasks.append(task)

        # We'll often end up with a task at the end that is the continuation with
        # no items or bindings, aside from a reference to the previous task, we
        # ignore those
        if not final_tasks or len(self.tasks[-1].bindings) or len(self.tasks[-1].items) > 1:
            final_tasks.append(self.tasks[-1])

        return final_tasks

#--------------------------------------------------
# Dataflow
#--------------------------------------------------

class Dataflow(c.Pass):
    def __init__(self, copying=True) -> None:
        super().__init__(copying)
        self.flow = Flow()

    def reset(self):
        super().reset()
        self.flow.reset()

    #--------------------------------------------------
    # Query
    #--------------------------------------------------

    def query(self, query: Task, parent=None):
        self.query_flow(query)
        if not parent:
            final_tasks = self.flow.finalize()
            query.behavior = Behavior.Sequence
            query.items = [build.relation_action(ActionType.Call, task, []) for task in final_tasks]

    def query_flow(self, task:Task, inline=False, ignore_deps=False, always_demand=False, raw_return=False):
        flow = self.flow
        flow.branch(ignore_deps=ignore_deps)

        for orig in task.items:
            ent:Any = orig.entity
            if orig.is_subtask_call():
                behavior = ent.value.behavior
                if behavior == Behavior.Union:
                    self.union_call(orig)
                elif behavior == Behavior.OrderedChoice:
                    self.ordered_choice_call(orig)
                elif behavior == Behavior.Query:
                    self.walk(orig, task)
            elif ent.isa(Builtins.Quantifier):
                self.quantifier_call(orig)
            elif ent.isa(Builtins.Aggregate):
                self.aggregate_call(orig)
            else:
                self.check_demand(orig, always_demand=always_demand, raw_return=raw_return)

        return flow.join()

    #--------------------------------------------------
    # Demand handling
    #--------------------------------------------------

    def check_demand(self, action:Action, always_demand=False, raw_return=False):
        flow = self.flow
        frame = self.flow.frame()

        if isinstance(action.entity.value, m.Property):
            frame.check_multi_prop(action.entity.value, action.params_list()[-1])

        # statics and always demands just get added directly
        if is_static(action) or always_demand:
            self.demand_action(action)

        # for non-relation gets (Types and Props) we need to store the demand
        elif action.action == ActionType.Get and not action.entity.isa(Builtins.Relation):
            self.store_demand(action)

        # Handle identity specially so we can break up big chains of adds
        elif action.entity.isa(Builtins.make_identity):
            self.demand_identity(action)

        # store construct calls
        elif action.action == ActionType.Construct:
            value = action.params_list()[-1]
            frame.push_demand(value, action)

        # For returns, we either split or keep it as is
        elif action.entity.isa(Builtins.Return):
            if not raw_return:
                self.split_return(action)
            else:
                self.demand_action(action)

        # For effects, we need to branch the flow so any demands don't kill the
        # original flow
        elif action.action.is_effect():
            flow.branch()
            self.demand_action(action)
            flow.join()
            flow.continue_()

        # For everything else, we just demand the vars and add the action
        else:
            self.demand_action(action)

    def store_demand(self, action:Action):
        flow = self.flow
        frame = flow.frame()
        root = action.entity.value
        if isinstance(root, m.Property):
            [key, value] = action.params_list()
            if is_static(value):
                self.demand_action(action)
            elif frame.has_demand(value) and frame.is_unique_demand(value, action):
                self.demand_action(action)
            elif frame.has_demand(key):
                self.demand(key)
                frame.push_demand(value, action, with_key=key)
            else:
                frame.push_demand(value, action, with_key=key)
        elif isinstance(root, m.Type):
            [key] = action.params_list()
            frame.push_demand(key, action)
            if not frame.key(key, None):
                frame.push_key(key, key)

    def demand_identity(self, action:Action):
        flow = self.flow
        frame = flow.frame()
        # if there are few vars leading into this add, we'll reify the identities
        # so that if there are many properties we only do the join once. 4 was chosen
        # randomly as a small arity
        parent_vars = gather_task_vars(frame.task, only_direct_children=True)
        flow.branch()
        cur_frame = flow.frame()
        self.demand_action(action)
        shared_vars = parent_vars & gather_task_vars(cur_frame.task, only_direct_children=True)
        vars = [*shared_vars, action.params_list()[-1]]
        if len(vars) <= 4:
            self.flow.append(build.relation_action(ActionType.Bind, cur_frame.task, vars))
            flow.join()
            flow.continue_()
            flow.frame().push_demand(vars[-1], build.relation_action(ActionType.Get, cur_frame.task, vars))
        else:
            # otherwise, just defer the identity demand and reify in each prop
            flow.join(discard=True)
            value = action.params_list()[-1]
            frame.push_demand(value, action)

    def demand_action(self, action:Action):
        for demand in action.vars(recursive=True):
            self.demand(demand)
        self.flow.append(action)

    def demand(self, var:Var, is_return=False, allow_missing=False):
        frame = self.flow.frame()
        for cur in frame.demand(var):
            if not frame.has_action(cur):
                if isinstance(cur.entity.value, m.Property):
                    [key, value] = cur.params_list()
                    if is_return and (allow_missing or frame.check_multi_prop(cur.entity.value)):
                        cur = build.relation_action(ActionType.Call, dsl.rel.pyrel_default._rel, [cur.entity.value, Builtins.Missing, key, value])
                    self.demand(key)
                elif cur.action == ActionType.Construct:
                    self.demand_action(cur)
                    continue
                elif isinstance(cur.entity.value, m.Type):
                    # For arbitrary relations (pyrel_default, make_identity), we need to demand keys
                    for key in cur.vars(recursive=True):
                        self.demand(key)
                self.flow.append(cur)

    #--------------------------------------------------
    # Return splitting
    #--------------------------------------------------

    def split_return(self, action:Action):
        flow = self.flow
        frame = flow.frame()

        all_keys = []
        hidden_keys = []
        unkeyed = []
        ret_params = action.params_list()
        is_distinct = action.entity.isa(Builtins.Distinct)

        available_roots = set()
        for ret in ret_params:
            if ret == frame.key(ret, None):
                available_roots.add(ret)

        for ret in ret_params:
            last_key = frame.key(ret, None)
            has_key = last_key and (not is_distinct or last_key in available_roots)

            if not has_key and last_key:
                if last_key not in hidden_keys:
                    hidden_keys.append(last_key)
                unkeyed.append(ret)
                self.demand(ret, is_return=True, allow_missing=True)
            elif has_key and last_key not in all_keys:
                all_keys.append(last_key)
            elif not last_key and ret.value is None:
                unkeyed.append(ret)

        multis = [x for x in ret_params if frame.is_multi(x)]
        if len(multis) > 1:
            all_keys.extend(multis)

        if len(unkeyed):
            ident = Var()
            flow.append(build.call(Builtins.make_identity, [Var(value=all_keys + unkeyed), ident]))
            all_keys.append(ident)

        for key in all_keys:
            self.demand(key, is_return=True)

        is_export = action.entity.isa(Builtins.ExportReturn)

        for (ix, ret) in enumerate(ret_params):
            flow.branch()
            self.demand(ret)
            col = Var(Builtins.Symbol, value=f"col{ix:03}")
            rel = dsl.rel.output.cols._rel
            if is_export:
                rel = dsl.rel.Export_Relation._rel
                uuid_str = dsl.rel.uuid_string._rel
                v = Var()
                flow.append(build.relation_action(ActionType.Call, dsl.rel.pyrel_default._rel, [uuid_str, ret, ret, v]))
                flow.append(build.relation_action(ActionType.Bind, rel, [col, *all_keys, v]))
            else:
                rel = dsl.rel.output.cols._rel
                flow.append(build.relation_action(ActionType.Bind, rel, [col, *all_keys, ret]))
            flow.join()

        return self.flow.continue_()

    #--------------------------------------------------
    # Union
    #--------------------------------------------------

    def union_call(self, call:Action):
        flow = self.flow

        # run through the subtasks and add them to the flow
        union_task = cast(Task, call.entity.value)
        union_task.parents.append(Builtins.InlineAnnotation)

        # find any vars that this union task depends on
        inner_vars = gather_task_vars(union_task)
        deps = list(flow.available_vars() & inner_vars)

        for item in union_task.items:
            neue = self.query_flow(cast(Task, item.entity.value), raw_return=True)
            rets = [i for i in neue.items if i.entity.isa(Builtins.Return)]
            if len(rets):
                rets[0].entity.value = union_task
                rets[0].action = ActionType.Bind
                if deps:
                    prepend_bindings(deps, rets[0])

        flow.continue_()
        # Get the result of the union
        get = build.relation_action(ActionType.Get, cast(Task, call.entity.value), deps + call.params_list())
        flow.append(get)

    #--------------------------------------------------
    # Ordered choice
    #--------------------------------------------------

    def ordered_choice_call(self, call:Action):
        flow = self.flow
        choice = cast(Task, call.entity.value)
        prevs = []
        branches = []

        # find any vars that this choice task depends on
        inner_vars = gather_task_vars(choice)
        deps = list(flow.available_vars() & inner_vars)

        for item in choice.items:
            cur_task = cast(Task, item.entity.value)

            # Negate all the previous branches to ensure we only return a value if
            # we're at our position in the order
            for prev in reversed(prevs):
                fetch = build.relation_action(ActionType.Get, prev, deps)
                prev_task = Task(items=[fetch])
                cur_task.items.insert(0, build.call(Builtins.Not, [Var(value=[]), Var(value=prev_task)]))

            neue:Task = self.query_flow(cur_task, raw_return=True)
            branches.append(neue)

            # find the return statement and turn it into a bind for the overall orderd_choice
            rets = [i for i in neue.items if i.entity.isa(Builtins.Return)]
            ret_bindings = [*deps]
            if len(rets):
                rets[0].entity.value = choice
                prepend_bindings(deps, rets[0])
            else:
                rets.append(build.relation_action(ActionType.Bind, choice, ret_bindings))
                neue.items.append(rets[0])

            # add a bind for this particular task so that we can negate it in subsequent
            # branches to create the ordering
            bind = build.relation_action(ActionType.Bind, neue, deps)
            neue.items.append(bind)

            prevs.append(neue)

        flow.continue_()
        # Get the result of the choice
        get = build.relation_action(ActionType.Get, choice, deps + call.params_list())
        flow.append(get)

    #--------------------------------------------------
    # Quantifiers
    #--------------------------------------------------

    def quantifier_call(self, call:Action):
        flow = self.flow
        quantifier = cast(Task, call.entity.value)
        group, task_var = [*call.bindings.values()]
        sub_task = self.query_flow(cast(Task, task_var.value), always_demand=True, ignore_deps=True)
        sub_task.parents.append(Builtins.InlineAnnotation)

        if isinstance(group.value, list) and len(group.value):
            raise Exception("TODO: grouped quantifiers")

        # Find any vars that this quantified task depends on
        sub_vars = gather_task_vars(sub_task)
        parent_vars = flow.available_vars()

        shared = sub_vars & parent_vars
        # bind those so we can use them in the quantified task
        sub_task.items.append(build.relation_action(ActionType.Bind, sub_task, shared))

        # create the quantified task, which just gets the subtask
        quantifed_task = Task()
        quantifed_task.items.append(build.relation_action(ActionType.Get, sub_task, shared))

        # add the call to the quantifier
        flow.continue_()
        for parent_var in parent_vars:
            self.demand(parent_var)
        flow.append(build.call(quantifier, [group, Var(value=quantifed_task)]))

    #--------------------------------------------------
    # Aggregates
    #--------------------------------------------------

    def aggregate_call(self, call:Action):
        flow = self.flow
        agg = cast(Task, call.entity.value)
        is_extender = agg.isa(Builtins.Extender)
        (args, group, ret) = call.params_list()
        group_vars = cast(List[Var], group.value)
        arg_vars = cast(List[Var], args.value)

        for demand in arg_vars + group_vars:
            self.demand(demand)

        # to prevent shadowing errors we need to map the inner vars to new vars
        mapped = [Var(name=var.name, type=var.type) for var in arg_vars]
        var_mapping = dict(zip(arg_vars, mapped))

        # create the inner relation we'll aggregate over
        flow.branch(inline=True, var_mapping=var_mapping)

        # vars that are in both the projection and grouping needed to be mapped in
        # the projection but made equivalent in the body so the grouping takes effect
        equivs = [(orig, neue) for (orig, neue) in var_mapping.items() if orig in group_vars]
        for (orig, neue) in equivs:
            flow.append(build.relation_action(ActionType.Call, Builtins.eq, [orig, neue]))
        # bind the mapped vars as the output of the inner relation
        for demand in arg_vars:
            self.demand(demand)
        flow.append(build.relation_action(ActionType.Bind, flow.frame().task, mapped))
        inner = flow.join()

        # create the outer aggregate
        flow.branch(ignore_deps=True)
        outer_bindings = [ret] if not is_extender else [ret, *arg_vars]
        outer_call = build.relation_action(ActionType.Call, agg, [Var(value=inner), *outer_bindings])
        flow.append(outer_call)

        out_params = [*group_vars, ret] if not is_extender else [ret, *group_vars, *arg_vars]
        flow.append(build.relation_action(ActionType.Bind, flow.frame().task, out_params))
        outer = flow.join()
        outer.parents.append(Builtins.InlineAnnotation)

        # Resume the flow
        flow.continue_()
        get = build.relation_action(ActionType.Get, outer, out_params)
        flow.append(get)


#--------------------------------------------------
# Shredder
#--------------------------------------------------

class Shredder(c.Pass):
    def query(self, query: Task, parent=None):
        neue_actions = []
        for item in query.items:
            if item.action not in [ActionType.Call, ActionType.Construct] and not item.entity.isa(Builtins.Relation):
                ident, action = item.entity, item.action
                for type in item.types:
                    neue_actions.append(build.relation_action(action, type, [ident]))
                for prop, value in item.bindings.items():
                    neue_actions.append(build.relation_action(action, prop, [ident, value]))
            else:
                walked = self.walk(item)
                neue_actions.append(walked)
        query.items = neue_actions

#--------------------------------------------------
# Splinter
#--------------------------------------------------

class Splinter(c.Pass):

    def query(self, query: Task, parent=None):
        effects = [i for i in query.items if i.action.is_effect()]

        grouped_effects = defaultdict(list)
        for item in effects:
            grouped_effects[(item.action, item.entity.value)].append(item)

        if len(grouped_effects) > 1:
            neue_items = []

            non_effects = [i for i in query.items if not i.action.is_effect()]
            effects_vars = gather_vars(effects)

            fetch = None
            if len(non_effects):
                fetch = self.create_fetch(non_effects, effects_vars)
                neue_items.append(fetch)

            for (k, b) in grouped_effects.items():
                neue_items.append(self.create_effect_query(b, effects_vars, fetch.entity.value if fetch else None))

            query.behavior = Behavior.Sequence
            query.items = neue_items

    #--------------------------------------------------
    # Subtask creation
    #--------------------------------------------------

    def create_fetch(self, non_effects: List[Action], effects_vars: Iterable[Var]):
        fetch = Task()
        fetch.parents.append(Builtins.InlineAnnotation)
        non_effects.append(build.relation_action(ActionType.Bind, fetch, effects_vars))
        fetch.items = non_effects
        return build.call(fetch, [])

    def create_effect_query(self, effects: List[Action], effects_vars: Iterable[Var], fetch: Any):
        neue = Task()
        if fetch:
            effects.insert(0, build.relation_action(ActionType.Get, fetch, effects_vars))
        neue.items = effects
        return build.call(neue, [])

#--------------------------------------------------
# SetCollector
#--------------------------------------------------

set_types = [ActionType.Bind, ActionType.Persist, ActionType.Unpersist]

class SetCollector(c.Pass):
    def query(self, query: Task, parent=None):
        binds = [i for i in query.items if i.action in set_types]
        if len(binds) > 1:
            neue_items = []
            for item in query.items:
                if item.action not in set_types:
                    neue_items.append(item)
            neue_items.extend(self.create_raw(binds))
            query.items = neue_items

    def create_raw(self, binds: List[Action]):
        vals = [Var(value=[]) for i in range(len(binds[0].bindings))]
        vars = [Var() for v in vals]

        for bind in binds:
            for ix, var in enumerate(bind.bindings.values()):
                cast(List[Var], vals[ix].value).append(var)

        return [
            build.relation_action(ActionType.Get, Builtins.RawData, vals + vars),
            build.relation_action(binds[0].action, cast(Type, binds[0].entity.value), vars)
        ]

#--------------------------------------------------
# Compiler
#--------------------------------------------------

class Clone(c.Pass):
    pass

class Compiler(c.Compiler):
    def __init__(self, config:config.Config):
        super().__init__(Emitter(), [
            Clone(),
            Shredder(),
            Dataflow(),
            Splinter(),
            SetCollector(),
        ])



from .clients import config as cfg
import ast
from collections import defaultdict
import io
import re
import textwrap
from typing import Any, List, Optional, Tuple
import warnings
from .metamodel import Action, Task
from . import debugging
from rich.console import Console
from rich.text import Text
from enum import Enum

#--------------------------------------------------
# Print helpers
#--------------------------------------------------

def rich_str(string:str, style:str|None = None) -> str:
    output = io.StringIO()
    console = Console(file=output, force_terminal=True)
    console.print(string, style=style)
    return output.getvalue()

def body_text(console, body:str):
    body = textwrap.dedent(body)
    for line in body.splitlines():
        if not line.startswith("  "):
            console.print(line)
        else:
            console.print(line, soft_wrap=True)

def mark_source(source, start_line:int, end_line:int, indent=8, highlight="yellow", highlight_lines = []):
    final_lines = []
    all_lines = source.source.splitlines()
    max_line = source.line + len(all_lines)
    line_number_len = len(str(max_line))
    for ix, line in enumerate(source.source.splitlines()):
        cur = line
        cur_indent = indent if ix > 0 else 0
        line_number = source.line + ix
        color = "dim white"
        if (line_number >= start_line and line_number <= end_line) or line_number in highlight_lines:
            color = highlight + " bold"
        cur = f"{' '*cur_indent}[{color}]  {line_number :>{line_number_len}} |  {cur}[/{color}]"
        final_lines.append(cur)
    return "\n".join(final_lines)

def print_source_error(source, name:str, content:str):
    fixed_content_length = len(name) + len(source.file) + len(str(source.line)) + 2  # 2 for the spaces around the dash
    num_dashes = 74 - fixed_content_length
    dashes = '-' * num_dashes
    console = Console(width=80, force_jupyter=False)
    console.print("\n")
    console.print(f"[red]--- {name} {dashes} {source.file}: {source.line}")
    console.print()
    body_text(console, content)
    console.print()
    console.print(f'[red]{"-" * 80}')
    console.print()

def print_error(name:str, content:str):
    fixed_content_length = len(name) + 2  # 2 for the spaces around the dash
    num_dashes = 76 - fixed_content_length
    dashes = '-' * num_dashes
    console = Console(width=80, force_jupyter=False)
    console.print("\n")
    console.print(f"[red]--- {name} {dashes}")
    console.print()
    body_text(console, content)
    console.print()
    console.print(f'[red]{"-" * 80}')
    console.print()

def print_error_name(name:str):
    console = Console(width=80, force_jupyter=False)
    console.print("\n")
    console.print(f'[red]{"-" * 80}')
    console.print()
    body_text(console, name)
    console.print()
    console.print(f'[red]{"-" * 80}')
    console.print()

#--------------------------------------------------
# Transformers
#--------------------------------------------------

class IfToWithTransformer(ast.NodeTransformer):
    def visit_If(self, node):
        with_node = ast.With(
            items=[ast.withitem(context_expr=node.test, ctx=ast.Load(), optional_vars=None)],
            body=node.body,
            lineno=node.lineno,
            type_comment=None)
        return with_node

class WithDynamic(ast.NodeTransformer):
    def visit_With(self, node):
        content = ast.unparse(node.items[0].context_expr).replace(")", "dynamic=True)")
        with_node = ast.With(
            items=[ast.withitem(context_expr=ast.Name(id=content), ctx=ast.Load(), optional_vars=node.items[0].optional_vars)],
            body=[],
            lineno=node.lineno,
            type_comment=None)
        return with_node

class SetToMethod(ast.NodeTransformer):
    def visit_Assign(self, node):
        if isinstance(node.targets[0], ast.Attribute):
            keyword = ast.keyword(arg=node.targets[0].attr, value=node.value)
            return ast.Expr(value=ast.Call(
                func=ast.Attribute(value=node.targets[0].value, attr="set", ctx=ast.Load()),
                args=[],
                keywords=[keyword],
                lineno=node.lineno,
                type_comment=None))
        return node

class AssignToCompare(ast.NodeTransformer):
    def visit_Assign(self, node):
        if isinstance(node.targets[0], ast.Attribute) and len(node.targets) == 1:
            compare_node = ast.Compare(
                left=node.targets[0],
                ops=[ast.Eq()],
                comparators=[node.value]
            )
            expr_node = ast.Expr(value=compare_node)
            ast.copy_location(expr_node, node)
            if hasattr(node, 'type_comment'):
                expr_node.type_comment = node.type_comment

            return expr_node
        return node

#--------------------------------------------------
# Finders
#--------------------------------------------------

class PropertyFinder(ast.NodeVisitor):
    def __init__(self, start_line, properties):
        self.errors = []
        self.start_line = start_line
        self.properties = properties
        self.found_properties_lines = []  # To store lines where properties are found
        self.dynamic_properties = []  # To store dynamic properties

    def to_line_numbers(self, node):
        return (node.lineno, node.end_lineno)

    def visit_Attribute(self, node):
        if node.attr in self.properties:
            line_numbers = self.to_line_numbers(node)
            if line_numbers[0] >= self.start_line:
                self.found_properties_lines.append(node.lineno)
        self.generic_visit(node)

    def visit_Call(self, node):
        # Check if this is a call to 'getattr'
        if (isinstance(node.func, ast.Name) and node.func.id == 'getattr' and
                len(node.args) >= 2):
            if isinstance(node.args[1], ast.Str):
                property_name = node.args[1].s
                if property_name in self.properties:
                    line_numbers = self.to_line_numbers(node)
                    if line_numbers[0] >= self.start_line:
                        self.found_properties_lines.append(node.lineno)
            else:
                line_numbers = self.to_line_numbers(node)
                if line_numbers[0] >= self.start_line:
                    self.dynamic_properties.append(node.lineno)
        self.generic_visit(node)

# --------------------------------------------------
# Exceptions
# --------------------------------------------------


class RAIException(Exception):
    def __init__(self, message, name=None, content=None, source=None, **kwargs):
        super().__init__(message)
        self.name = name
        self.message = message
        self.content = content
        self.source = source
        self.raw_content = self.strip_rich_tags()

        # Store any additional keyword arguments as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        return self.message

    def pprint(self):
        if self.source and self.content:
            print_source_error(self.source, self.name or self.message, self.content.strip())
        elif self.content:
            print_error(self.name or self.message, self.content.strip())
        else:
            print_error_name(self.name or self.message)

    def __eq__(self, other):
        if isinstance(other, RAIException):
            return self.message == other.message and self.__dict__ == other.__dict__
        return False

    def strip_rich_tags(self):
        if self.content:
            # Use Text.from_markup to remove rich tags
            plain_text = Text.from_markup(self.content).plain
            # Split the content into lines
            lines = plain_text.split('\n')
            # Remove leading empty lines or lines with only whitespace
            while lines and not lines[0].strip():
                lines.pop(0)
            # Strip leading whitespace from the first non-empty line
            if lines:
                lines[0] = lines[0].lstrip()
            # Join the lines back together
            cleaned_content = '\n'.join(lines)
            return cleaned_content
        else:
            return self.message

    def clone(self):
        return RAIException(message=self.message, name=self.name, content=self.content, source=self.source)

class RelQueryError(Exception):
    def __init__(self, problems):
        super().__init__("Rel query error")
        self.problems = problems

    def pprint(self):
        body = "\n\n".join(
            f'{problem["error_code"]}\n{problem["report"]}\n{problem["message"]}'
            for problem in self.problems
        )
        return f"Rel query error\n{body}"

    def clone(self):
        return RelQueryError(self.problems)


# --------------------------------------------------
# Rel Exceptions
# --------------------------------------------------
class UninitializedPropertyException(RAIException):
    def __init__(self, undefineds: List[Tuple[str, Any]]):
        self.content = ""
        self.raw_content = ""
        self.source = None
        self.name = "Uninitialized property"
        self.message = "Uninitialized property"

        # group by source
        source_map = defaultdict(list)
        for name, source in undefineds:
            source_map[source.source].append((source, name))

        for names in source_map.values():
            self.source = names[0][0]
            unique_names = list(set([name for (_, name) in names]))
            props = ", ".join([f"[yellow]{name}[/yellow]" for name in unique_names])
            prop_line = (
                f"property {props} has"
                if len(unique_names) == 1
                else f"properties {props} have"
            )
            found = PropertyFinder(source.line, unique_names)
            if source.block:
                found.visit(source.block)
            found_lines = found.found_properties_lines or found.dynamic_properties
            marked = mark_source(source, -1, -1, indent=12, highlight_lines=found_lines)
            self.content = textwrap.dedent(f"""
            The {prop_line} never been set or added to and so will always cause the rule or query to fail.

            {marked}
            """)

        super().__init__(self.name, self.message, self.content, self.source)


class RAITypeError(RAIException, TypeError):
    def __init__(self, param: str, message: str):
        self.param = param
        self.message = message
        self.name = "Invalid type error"
        self.source = Errors.call_source(4)

        # Mark the source if available
        if self.source:
            self.marked = mark_source(self.source, self.source.line, self.source.line)
        else:
            self.marked = ""

        # Format the content and raw_content
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()

        super().__init__(self.message, self.name, self.content, self.source)

    def format_message(self):
        if self.source:
            return textwrap.dedent(f"""
            The parameter [yellow]{self.param}[/yellow] is of the wrong type.

            {self.marked}
            """)
        else:
            return textwrap.dedent(f"""
            The parameter [yellow]{self.param}[/yellow] is of the wrong type.
            """)


# --------------------------------------------------
# DSL scope errors
# --------------------------------------------------
class OutOfContextException(RAIException):
    def __init__(self, source):
        self.name = "Outside of context"
        self.message = (
            "Looks like this object is being used outside of a rule or query."
        )
        self.source = source
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content, source=source)

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        Looks like this [yellow]object[/yellow] is being used outside of a rule or query.

        {marked}
        """)


class VariableOutOfContextException(RAIException):
    def __init__(self, source, name: str, is_property=False):
        self.name = "Variable out of context"
        self.message = f"Looks like a variable representing '{name}' is being used outside of the rule or query it was defined in."
        self.source = source
        self.is_property = is_property
        self.content = self.format_message(name)

        super().__init__(self.message, self.name, self.content, source=source)

    def format_message(self, name):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        Looks like a variable representing [yellow bold]{name}[/yellow bold] is being used outside of the rule or query it was defined in.

        {marked}
        """)


# --------------------------------------------------
# DSL property exceptions
# --------------------------------------------------


class ReservedPropertyException(RAIException):
    def __init__(self, source, property_name: str):
        self.name = "Reserved property name"
        self.message = f"The property '{property_name}' is a reserved property name on RelationalAI types."
        self.source = source
        self.content = self.format_message(property_name)

        super().__init__(self.message, self.name, self.content, source=self.source)

    def format_message(self, property_name):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        The property '{property_name}' is a reserved property name on RelationalAI types.

        {marked}
        """)


class NonCallablePropertyException(RAIException):
    def __init__(self, source, property_name: str):
        self.name = "Non-callable property"
        self.message = f"The property '{property_name}' is not callable."
        self.source = source
        self.content = self.format_message(property_name)

        super().__init__(self.message, self.name, self.content, source=source)

    def format_message(self, property_name):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        The property '{property_name}' is not callable on RelationalAI types.

        {marked}
        """)

class InvalidPropertySetException(RAIException):
    def __init__(self, source):
        self.name = "Invalid property set"
        self.message = "You can't set properties directly on a RAI object."
        self.source = source
        self.start_line = source.line
        self.end_line = source.line
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()
        super().__init__(
            message=self.message, name=self.name, source=source, content=self.content
        )

    def format_message(self):
        marked = self.mark_source()
        dynamic = mark_source(
            self.source.modify(SetToMethod()),
            self.start_line,
            self.end_line,
            highlight="green",
        )
        compare = mark_source(
            self.source.modify(AssignToCompare()),
            self.start_line,
            self.end_line,
            highlight="green",
        )

        return textwrap.dedent(f"""
        You can't set properties directly on a RAI object.

        {marked}

        If you are trying to set the value of a property use [green]set()[/green]:

        {dynamic}

        Or maybe you meant [green]==[/green] instead?

        {compare}
        """)

    def mark_source(self):
        return mark_source(self.source, self.start_line, self.end_line)


# --------------------------------------------------
# Engine exceptions
# --------------------------------------------------


class EngineNotFoundException(RAIException):
    def __init__(self, engine_name: str, message: str):
        self.engine_name = engine_name
        self.message = message
        self.name = "Engine unavailable"

        # Format the content and raw_content
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        return textwrap.dedent(f"""
        The engine [yellow]{self.engine_name}[/yellow] isn't available. You can start the engine with the following command:

        [green]rai engines:create --name {self.engine_name}[/green]
        """)

class EnginePending(RAIException):
    def __init__(self, engine_name: str):
        self.engine_name = engine_name
        self.message = "Engine is in a pending state"
        self.name = "Engine not ready"

        # Format the content and raw_content
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        return textwrap.dedent(f"""
        The engine [yellow]{self.engine_name}[/yellow] is in a pending state. You can see the status of engines with the following command:

        [green]rai engines:list[/green]
        """)


# --------------------------------------------------
# Snowflake errors
# --------------------------------------------------
class SnowflakeAppMissingException(RAIException):
    def __init__(self, app_name: str):
        self.app_name = app_name
        self.message = f"The snowflake app '{app_name}' doesn't appear to be installed in this snowflake account."
        self.name = "Couldn't find RelationalAI snowflake application"
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        return textwrap.dedent(f"""
        The snowflake app '{self.app_name}' doesn't appear to be installed in this snowflake account.

        If it's installed under a different name, run 'rai init' on the command line and you can set the app name.
        """)


class SnowflakeImportMissingException(RAIException):
    def __init__(self, source: str, import_name: str, model_name: str):
        self.source = source
        self.import_name = import_name
        self.model_name = model_name
        self.message = (
            f"The Snowflake object '{import_name}' hasn't been imported into RAI."
        )
        self.name = "Couldn't find import"
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        return textwrap.dedent(f"""
        The Snowflake object '{self.import_name}' hasn't been imported into RAI.

        You can create an import for it using the rai cli:

          [green]rai imports:stream --source {self.import_name} --model {self.model_name}[/green]
        """)


class SnowflakeChangeTrackingNotEnabledException(RAIException):
    def __init__(self, obj: str, sql: str):
        self.obj = obj
        self.sql = sql
        self.message = f"Change tracking isn't enabled for '{obj}'."
        self.name = "Change tracking not enabled"
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        return textwrap.dedent(f"""
        Change tracking isn't enabled for [yellow]{self.obj}[/yellow].

        To enable change tracking, you'll need to run the following SQL:

        [green]{self.sql}[/green]
        """)


class SnowflakeInvalidSource(RAIException):
    def __init__(self, source, type_source: str):
        self.source = source
        self.type_source = type_source
        self.message = "Invalid source"
        self.name = "Invalid source"
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content, self.source)

    def format_message(self):
        marked = mark_source(self.source, self.source.line, self.source.line)
        return textwrap.dedent(f"""
        The source provided isn't a fully qualified Snowflake table or view name:

        {marked}
        """)


class SnowflakeRaiAppNotStarted(RAIException):
    def __init__(self, app_name: str, sql: str):
        self.app_name = app_name
        self.sql = sql
        self.message = f"The RelationalAI app '{app_name}' isn't started."
        self.name = "The RelationalAI application not started"
        self.content = self.format_message()

        super().__init__(self.message, self.name, self.content)

    def format_message(self):
        return textwrap.dedent(f"""
        Your app '{self.app_name}' in the RelationalAI service isn't started. 
        You can start it by running the following SQL:

        [green]{self.sql}[/green]
        """)


# --------------------------------------------------
# RAI Warnings
# --------------------------------------------------
class RAIWarning(Warning):
    def __init__(
        self,
        name,
        source: Optional[any] = None,
        content: Optional[str] = None,
    ):
        self.name = name
        self.source = source
        self.content = content
        self.raw_content = self.strip_rich_tags()

        self.trigger_warning()
        warnings.warn(self)

    def pprint(self):
        if self.source and self.content:
            print_source_error(self.source, self.name or self.message, self.content)
        elif self.content:
            print_error(self.name or self.message, self.content)
        else:
            print_error_name(self.name or self.message)

    def trigger_warning(self):
        from .debugging import warn

        warn(self)

    def strip_rich_tags(self):
        if self.content:
            # Use Text.from_markup to remove rich tags
            plain_text = Text.from_markup(self.content).plain
            # Split the content into lines
            lines = plain_text.split('\n')
            # Remove leading empty lines or lines with only whitespace
            while lines and not lines[0].strip():
                lines.pop(0)
            # Strip leading whitespace from the first non-empty line
            if lines:
                lines[0] = lines[0].lstrip()
            # Join the lines back together
            cleaned_content = '\n'.join(lines)

            return cleaned_content
        else:
            return self.name

    def __str__(self):
        return f"{self.name}: {self.raw_content}"


class InvalidIfWarning(RAIWarning):
    def __init__(self, task: Task | Action, start_line: int, end_line: int):
        self.name = "Invalid if"
        self.source = debugging.get_source(task)
        self.start_line = start_line
        self.end_line = end_line
        self.content = self.format_message()

        super().__init__(
            name=self.name,
            source=self.source,
            content=self.content,
        )

    def format_message(self):
        marked = self.mark_source()
        updated = mark_source(
            self.source.modify(IfToWithTransformer()),
            self.start_line,
            self.end_line,
            highlight="green",
        )
        dynamic = mark_source(
            self.source.modify(WithDynamic()),
            self.source.line,
            self.end_line,
            highlight="green",
        )

        return textwrap.dedent(f"""
        In a RelationalAI query, using an if statement dynamically modifies the structure of the query itself, rather than adding a conditional.

        {marked}

        If you're trying to do an action based on a condition, use a [green]with[/green] statement instead.

        {updated}

        If you are trying to create a dynamic query where parts are conditional, add the [green]dynamic=True[/green] flag to the query like so:

        {dynamic}
        """)

    def mark_source(self):
        return mark_source(self.source, self.start_line, self.end_line)


class InvalidLoopWarning(RAIWarning):
    def __init__(self, task: Task | Action, start_line: int, end_line: int):
        self.name = "Invalid loop"
        self.source = debugging.get_source(task)
        self.start_line = start_line
        self.end_line = end_line
        self.content = self.format_message()
        super().__init__(
            name=self.name,
            source=self.source,
            content=self.content,
        )

    def format_message(self):
        marked = self.mark_source()
        dynamic = mark_source(
            self.source.modify(WithDynamic()),
            self.source.line,
            self.end_line,
            highlight="green",
        )

        return textwrap.dedent(f"""
        In a RelationalAI query, using a loop statement would dynamically modify the query itself, like a macro.

        {marked}

        If that's the goal, you can add the [green]dynamic=True[/green] flag to the query:

        {dynamic}
        """)

    def mark_source(self):
        return mark_source(self.source, self.start_line, self.end_line)


class InvalidTryWarning(RAIWarning):
    def __init__(self, task: Task | Action, start_line: int, end_line: int):
        self.name = "Invalid try"
        self.source = debugging.get_source(task)
        self.start_line = start_line
        self.end_line = end_line
        self.content = self.format_message()

        super().__init__(
            name=self.name,
            source=self.source,
            content=self.content,
        )

    def format_message(self):
        marked = self.mark_source()
        dynamic = mark_source(
            self.source.modify(WithDynamic()),
            self.source.line,
            self.end_line,
            highlight="green",
        )

        return textwrap.dedent(f"""
        In a RelationalAI query, using a try statement will have no effect unless a macro-like function is being called and can fail.

        {marked}

        If macro-like behavior is the goal, you can add the [green]dynamic=True[/green] flag to the query:

        {dynamic}
        """)

    def mark_source(self):
        return mark_source(self.source, self.start_line, self.end_line)


class InvalidBoolWarning(RAIWarning):
    def __init__(self, source):
        self.name = "Invalid boolean expression with Producer"
        self.source = source
        self.start_line = source.line
        self.end_line = source.line
        self.content = self.format_message()
        self.raw_content = self.strip_rich_tags()
        super().__init__(self.name, source=source, content=self.content)

    def format_message(self):
        marked = self.mark_source()
        return textwrap.dedent(f"""
        In a RelationalAI query, the truth values of Producer objects are unknown
        until the query has been evaluated. You may not use Producers in boolean
        expressions, such as those involving [green]if[/green], [green]while[/green], [green]and[/green], [green]or[/green], and [green]not[/green]:

        {marked}

        Producer objects include:

        - [green]Instance[/green] objects, such as [green]person[/green].
        - [green]InstanceProperty[/green] objects, such as [green]person.age[/green].
        - [green]Expression[/green] objects, such as [green]person.age >= 18[/green].
        """)

    def mark_source(self):
        return mark_source(self.source, self.start_line, self.end_line)

# whether the python library or the Rel schema is out of date (or both!)
class InvalidVersionKind(Enum):
    LibraryOutOfDate = 'LibraryOutOfDate'
    SchemaOutOfDate = 'SchemaOutOfDate'
    Incompatible = 'Incompatible'

class RAIInvalidVersionWarning(RAIWarning):
    def __init__(
        self,
        kind: InvalidVersionKind,
        expected: List[Tuple[str, str, str]],
        lock: dict,
        platform: str,
        errors: List[Tuple[InvalidVersionKind, str, str]] = [],
        config: cfg.Config = None,
        model_name: str = "",
    ):

        self.libraries = self.get_current_libs(errors)
        self.expected_lib_version = self.get_expected_lib_version(expected)
        self.current_lib_version = self.get_current_lib_version(lock)
        self.config = config
        self.model_name = model_name

        self.name = "Invalid version"

        if kind == InvalidVersionKind.LibraryOutOfDate:
            msg = textwrap.dedent("""
                The RelationalAI python library is out of date with respect to the schema. You can update the library using the following command:

                [green]pip install relationalai --upgrade[/green]
            """).strip()
        elif kind == InvalidVersionKind.SchemaOutOfDate:
            schema_fix = textwrap.dedent(f"""
                You can update the schema using the following stored procedure:

                {self.generate_command_line()}
                """).strip()

            if platform == "azure":
                schema_fix = "Unfortunately, the schema must be recreated in Azure."
            schema_msg = textwrap.dedent("""
                The schema is out of date with respect to the RelationalAI library.
                """).strip()
            msg = f"{schema_msg} {schema_fix}"
        else:
            lib_msg = textwrap.dedent("""
                The RelationalAI python library is incompatible with the current schema.
                You can update the library using the following command:

                [green]pip install relationalai --upgrade[/green]
            """).strip()
            if platform == "azure":
                schema_fix = "In addition, the schema must be recreated in Azure."
            else:
                schema_fix = textwrap.dedent(f"""
                In addition, you can update the schema using the following stored procedure:

                {self.generate_command_line()}
                """).strip()
            msg = f"{lib_msg}\n\n{schema_fix}"

        if (len(self.libraries) > 0):
            library_info = self.generate_library_info()
            full_msg = f"{msg}\n\n{library_info}"
        else:
            full_msg = msg

        super().__init__(name=self.name, content=full_msg)

    def extract_version(self, version_obj):
        match = re.search(r"\('([^']+)'\)", repr(version_obj))
        if match:
            return match.group(1)
        else:
            return ""

    def get_expected_lib_version(self, expected):
        libraries = {}
        for name, _, version_range in expected:
            from_version = self.extract_version(version_range[0])
            to_version = self.extract_version(version_range[1])
            libraries[name] = {"from": from_version, "to": to_version}
        return libraries

    def get_current_lib_version(self, lock):
        libraries = {}
        for k, v in lock.items():
            libraries[k[0]] = v
        return libraries

    def get_current_libs(self, errors=[]):
        if len(errors) == 0:
            return []

        libraries = []
        for _, library, _ in errors:
            libraries.append(library)
        return libraries

    def generate_library_info(self):
        if len(self.libraries) == 1:
            lib = self.libraries[0]
            current_version = self.current_lib_version.get(lib, "unknown")
            expected_version = self.expected_lib_version.get(lib, {})
            from_version = expected_version.get("from", "unknown")
            to_version = expected_version.get("to", "unknown")
            return (
                f"Your current library '{lib}' has version {current_version}. "
                f"However, it is expected to have a version range between {from_version} and {to_version}."
            )
        else:
            info = "Your current libraries have the following issues:\n"
            for lib in self.libraries:
                current_version = self.current_lib_version.get(lib, "unknown")
                expected_version = self.expected_lib_version.get(lib, {})
                from_version = expected_version.get("from", "unknown")
                to_version = expected_version.get("to", "unknown")
                info += (
                    f"- Library '{lib}' has version {current_version}. "
                    f"Expected version range: {from_version} to {to_version}.\n"
                )
            return info.strip()

    def generate_command_line(self):
        rai_app = self.config.get("rai_app_name", "") if self.config else ""
        engine_name = self.config.get("engine", "") if self.config else ""
        model_name = self.model_name
        if rai_app and engine_name and model_name:
            command = f"[green]CALL {rai_app}.API.update_libraries('{model_name}', '{engine_name}')[/green]"
        else:
            command = "[green]CALL API.update_libraries()[/green]"
        return command


#TODO: Take out call_source from Errors class and put it in the main module
class Errors:
    @staticmethod
    def call_source(steps=None):
        if steps is None:
            return debugging.capture_code_info()
        return debugging.capture_code_info(steps + 1)



from relationalai.clients.util import scrub_exception

# contents: `{"relation_count": 1}` in an internal account
url = "https://sfc-prod3-ds1-27-customer-stage.s3.us-west-2.amazonaws.com/z00j0000-s/stages/c0e19bda-0e9b-47b2-a21d-cfcdd5a16945/transaction-results/01b458ec-0002-6d1e-0051-c00704e7a71a/relation-count.json?X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAXYKJQIRCBD5PYLZS%2F20240515%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20240515T154106Z&X-Amz-Expires=3600&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEEAaCXVzLXdlc3QtMiJGMEQCIC8EHn51SngV94nYpQBrKBPZ%2B%2Fiifzpj3u15FAMFyM5rAiBSsNFEILg7%2FpucuFhcYMsjAGIpZADIzlpemCEsNZBBjCrJAwip%2F%2F%2F%2F%2F%2F%2F%2F%2F%2F8BEAAaDDUzMzI2Njk3Mzc2NCIMxWhZCLdXnNlNL8YMKp0Do8jSddBWhvb4Z4IBXWcJ7vjuGbai418ToGUAAS9%2B5Kq7j0%2BP7Lbsh0mZ87VtDf0FXjkYyW4fD1flAB%2Fi0MFeGfPOkAoC7l0aBx1aOAE39iAGL5EgNDvKubUt7i3cWPe7wiubDrZMKhV6yoeVMjHZu2rOsoYl%2F2A2gfK%2By8K2Kvojlbfsx3%2Bie%2FuHNSVC%2FrUif%2BZuPcNJjtXxjpHb2LxZcyxnVYCDChdCtFG4%2BPsG55woeGJXgUQGSF3g8fdEZ7XckTpbn%2FE9leGOEqourELgjJdPpwXzjXZYs2y%2FFtxFuSOofghi9B9Lh9gfcKuWTfKhiambpz9hm1oN8caeUVnZG1o0c8VZ0ZecBY84elMv%2F5PElno4unDTnSmCqDB7aMdMlS6i9sO0J5NKnJ4HUjR%2BHT7UuOAljBlT1b2jSvPQncOkhtLila7hYeq1X1YVZNg7SkwI3ZZz5y5Xu0kWjLKR9tElN5%2B%2BoukHxtLYH7wU6FRMGtcyW%2BtPAH33VxQX6BQ7Fy4%2B6veydANT%2FaAp8XiB5dvEp6Kz5pEDOkm61u8w5a6TsgY6mgHMzNSd8ffllHz6PUsXLlgalp%2FXWnu%2B9aUy1dGMT46jFEfhJrwq5p48R7PE9qKYx0rjPuO5z5hkI2OZlu2iFTJ02biWuR2AzGC3becjqBsA0l2Dcsdp5U5U0GmSPAgL4mByhGsKvVB5i6Qj08irY5%2FPzAdZz4Nhy68sm8W81bgLNdpk7vGEx5Q8eNGX3P3DiGgHd82dVK9hrfa1&X-Amz-SignedHeaders=host&X-Amz-Signature=6145b734ae683536221b4d5d5cc379df3cc7211e4bd72381b88fcbb2eaf78dc4"

def test_scrub_exceptions():
    # exception with scrubbed URL
    exc = Exception(f"connection refused to {url}")
    scrubbed_exc = scrub_exception(exc)
    assert scrubbed_exc.message == "connection refused to https://sfc-prod3-ds1-27-customer-stage.s3.us-west-2.amazonaws.com/z00j0000-s/stages/c0e19bda-0e9b-47b2-a21d-cfcdd5a16945/transaction-results/01b458ec-0002-6d1e-0051-c00704e7a71a/relation-count.json?X-Amz-Algorithm=XXX&X-Amz-Credential=XXX&X-Amz-Date=XXX&X-Amz-Expires=XXX&X-Amz-Security-Token=XXX&X-Amz-SignedHeaders=XXX&X-Amz-Signature=XXX"
    
    exc2 = Exception("some other error")
    scrubbed_exc2 = scrub_exception(exc2)
    assert scrubbed_exc2 == exc2



import inspect
import os
import pickle
import sys
from typing import Any, Callable, Type
import hypothesis
from hypothesis.core import encode_failure
from hypothesis.errors import DidNotReproduce
from relationalai import metamodel as mm
from gentest.util import PROJECT_DIR
from gentest.harness.database import HYPO_DB_PATH, expand_short_path, get_conjecture_data
from gentest.validate.errors import GenException
from gentest.cli.collect_tests import collect_tests, pytest_node_fn, pytest_node_name


from gentest.emit import DocumentEmitter

def get_fn_file(fn, relative_to = os.path.join(PROJECT_DIR, "..")):
    file = sys.modules[fn.__module__].__file__
    return os.path.relpath(file, relative_to) if file else None

def get_test_for(repro_path: str):
    if HYPO_DB_PATH  in repro_path:
        repro_path = os.path.relpath(repro_path, HYPO_DB_PATH)
    key_hash = repro_path[0:repro_path.find("/")]
    tests = collect_tests()
    for test in tests:
        fn = pytest_node_fn(test)
        cur_key = getattr(fn, "__key_hash", "")
        if cur_key.startswith(key_hash):
            break
    else:
        raise Exception("Unable to find a matching test for the given key hash.")

    return (test, fn)


def get_failure_repro_blob(conjecture_data: bytes):
    return encode_failure(conjecture_data)

def get_generated_example_for_fn(blob: bytes, fn: Any) -> tuple[tuple, list[Any]]:
    params = inspect.signature(fn.inner.hypothesis.inner_test).parameters
    types = [param.annotation if param.annotation else "Any" for param in params.values() if param.name != "data"]

    args = ()
    def collect_args(**kwargs):
        nonlocal args
        args = tuple(kwargs[param] for param in params.keys() if param != "data")

    prev_inner = fn.inner.hypothesis.inner_test
    fn.inner.hypothesis.inner_test = collect_args
    try:
        fn()
    except Exception:
        pass
    finally:
        fn.inner.hypothesis.inner_test = prev_inner

    return (args, types)

def encode(short_path: str):
    repro_path = expand_short_path(short_path, HYPO_DB_PATH)
    print(f"Retrieving repro input from {os.path.relpath(repro_path, os.path.dirname(PROJECT_DIR))}")
    data = get_conjecture_data(repro_path)
    blob = get_failure_repro_blob(data)
    print(f"Reproduce the failure by adding the following decorator to the test case: `@reproduce_failure(\"{hypothesis.__version__}\", {repr(blob)})`")

Shareable = tuple[str, str, bytes]

def encode_shareable(short_path: str):
    repro_path = expand_short_path(short_path, HYPO_DB_PATH)
    print(f"Retrieving repro input from {os.path.relpath(repro_path, os.path.dirname(PROJECT_DIR))}")
    data = get_conjecture_data(repro_path)
    blob = get_failure_repro_blob(data)
    test, fn = get_test_for(repro_path)
    return (get_fn_file(fn.inner), pytest_node_name(test), blob)

def reproduce(short_path: str):
    repro_path = expand_short_path(short_path, HYPO_DB_PATH)
    print(f"Retrieving repro input from {os.path.relpath(repro_path, os.path.dirname(PROJECT_DIR))}")
    data = get_conjecture_data(repro_path)
    blob = get_failure_repro_blob(data)
    test, fn = get_test_for(repro_path)
    print(f"Reproducing failure for test {pytest_node_name(test)} in {get_fn_file(fn.inner)} with input {repr(blob)}")
    return reproduce_blob(fn, blob)

def reproduce_blob(test_fn: Callable, blob: bytes):
    inner = test_fn.inner
    try:
        test_fn.inner = hypothesis.reproduce_failure(hypothesis.__version__, blob)(test_fn.inner)
        test_fn(False)
    except GenException as err:
        print(err.pprint())
    except DidNotReproduce:
        print()
        print("✨ Could not reproduce failure. You fixed it! ✨")
    finally:
        test_fn.inner = inner
        del test_fn.inner._hypothesis_internal_use_reproduce_failure

def get_example_inputs(short_path: str):
    repro_path = expand_short_path(short_path, HYPO_DB_PATH)
    print(f"Retrieving repro input from {os.path.relpath(repro_path, os.path.dirname(PROJECT_DIR))}")
    data = get_conjecture_data(repro_path)
    blob = get_failure_repro_blob(data)
    test, fn = get_test_for(repro_path)
    print(f"Generating input for {pytest_node_name(test)} in {get_fn_file(fn.inner)} from {repr(blob)}")
    args, types = get_generated_example_for_fn(blob, fn)
    return args, types

def get_example(short_path: str, ix = 0) -> tuple[Any, Type]:
    inputs = get_example_inputs(short_path)
    return inputs[0][ix], inputs[1][ix]

def export_inputs(short_path: str, out: str):
    args, types = get_example_inputs(short_path)
    print(f"Writing repro to {out} as tuple of ({', '.join(t.__name__ for t in types)})")
    with open(out, "wb") as file:
        pickle.dump(args, file)

def export_ir(short_path: str, out: str, example: tuple[Any, Type]|None = None, ix = 0):
    if not example:
        example = get_example(short_path, ix)
    value, arg_type = example
    print(f"Writing {arg_type.__name__} to {out}")
    with open(out, "wb") as file:
        pickle.dump(value, file)

def emit_task(task: mm.Task):
    doc = DocumentEmitter()
    doc.rule(task)
    return doc.stringify().strip()

EMITTERS = {
    mm.Task: emit_task
}

def export_pyrel(short_path: str, out: str, example: tuple[Any, Type]|None = None, ix = 0):
    if not example:
        example = get_example(short_path, ix)
    value, arg_type = example
    print(f"Emitting pyrel for {arg_type.__name__}")
    if arg_type not in EMITTERS:
        raise Exception(f"Failed to emit pyrel for {arg_type.__name__}, no registered emitter.")

    emitter = EMITTERS[arg_type]

    emitted = emitter(value)
    print(f"Writing pyrel to {out}")
    with open(out, "w") as file:
        file.write(emitted)



#pyright: reportUnusedExpression=false
import relationalai as rai

model = rai.Model(name=globals().get("name", "test_bool"), config=globals().get("config"))
Person = model.Type("Person")
Adult = model.Type("Adult")

with model.rule():
    Person.add(name="Joe", age=74)
    Person.add(name="Bob", age=40)
    Person.add(name="Jane", age=10)

with model.rule():
    p = Person()
    p.age >= 18
    p.set(Adult, cool=False)

with model.rule():
    p = Person()
    p.age < 18
    p.set(cool=True)

with model.query() as select:
    a = Adult()
    a.cool == False # noqa
    z = select(a, a.name, a.age)

print(z.results)

with model.query() as select:
    p = Person(cool=True)
    z = select(p, p.name)

print(z.results)



#pyright: reportPrivateImportUsage=false

import io
import os
from re import Pattern
from typing import Sequence, cast, Any, List
from pathlib import Path
from InquirerPy.base.complex import FakeDocument
from prompt_toolkit.key_binding import KeyPressEvent
from prompt_toolkit.validation import ValidationError
import rich
from rich.console import Console
import sys
from InquirerPy import inquirer, utils as inquirer_utils
from InquirerPy.base.control import Choice
import time
import threading
import itertools
from ..debugging import jupyter

#--------------------------------------------------
# Constants
#--------------------------------------------------

REFETCH = "[REFETCH LIST]"

#--------------------------------------------------
# Style
#--------------------------------------------------

STYLE = inquirer_utils.get_style({
    "fuzzy_prompt": "#e5c07b"
}, False)

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def rich_str(string:str, style:str|None = None) -> str:
    output = io.StringIO()
    console = Console(file=output, force_terminal=True)
    console.print(string, style=style)
    return output.getvalue()

def nat_path(path: Path, base: Path):
    resolved_path = path.resolve()
    resolved_base = base.resolve()
    if resolved_base in resolved_path.parents or resolved_path == resolved_base:
        return resolved_path.relative_to(resolved_base)
    else:
        return resolved_path.absolute()

def get_default(value:str|None, list_of_values:Sequence[str]):
    if value is None:
        return None
    list_of_values_lower = [v.lower() for v in list_of_values]
    value_lower = value.lower()
    if value_lower in list_of_values_lower:
        return value

#--------------------------------------------------
# Dividers
#--------------------------------------------------

def divider(console=None, flush=False):
    div = "\n[dim]---------------------------------------------------\n "
    if console is None:
        rich.print(div)
    else:
        console.print(div)
    if flush:
        sys.stdout.flush()

def abort():
    rich.print()
    rich.print("[yellow]Aborted")
    divider()
    sys.exit(1)

#--------------------------------------------------
# Prompts
#--------------------------------------------------

default_bindings = cast(Any, {
    "interrupt": [
        {"key": "escape"},
        {"key": "c-c"},
        {"key": "c-d"}
    ],
    "skip": [
        {"key": "c-s"}
    ]
})

def prompt(message:str, value:str|None, newline=False, validator:callable = None, invalid_message:str|None = None) -> str:
    if value:
        return value
    try:
        result:str = inquirer.text(
            message,
            validate=validator,
            invalid_message=invalid_message,
            keybindings=default_bindings,
        ).execute()
    except KeyboardInterrupt:
        abort()
        raise Exception("Unreachable")
    if newline:
        rich.print("")
    return result

def select(message:str, choices:List[str|Choice], value:str|None, newline=False, **kwargs) -> str|Any:
    if value:
        return value
    try:
        result:str = inquirer.select(message, choices, keybindings=default_bindings, **kwargs).execute()
    except KeyboardInterrupt:
        abort()
        raise Exception("Unreachable")
    if newline:
        rich.print("")
    return result

def fuzzy(message:str, choices:List[str], default:str|None = None, multiselect=False, show_index=False, **kwargs) -> str:
    for i, choice in enumerate(choices):
        if isinstance(choice, str):
            choices[i] = {"name": f"{i+1} {choice}" if show_index else choice, "value": choice}

    values = [choice["value"] for choice in choices]

    try:
        kwargs["keybindings"] = default_bindings
        if multiselect:
            kwargs["keybindings"] = {
                "toggle": [
                    {"key": "tab"},   # toggle choices
                ],
                "toggle-down": [
                    {"key": "tab", "filter":False},
                ],
            }.update(default_bindings)
            kwargs["multiselect"] = True

        scroll_to_ix = None
        # @FIXME: Future improvement to make it optionally case insensitive
        if default and default in values:
            scroll_to_ix = values.index(default)
            default = None

        prompt = inquirer.fuzzy(message, choices=choices, default=default or "", max_height=8, border=True, style=STYLE, **kwargs)
        if scroll_to_ix is not None:
            prompt.content_control.selected_choice_index = scroll_to_ix
        return prompt.execute()
    except KeyboardInterrupt:
        return abort()

def fuzzy_with_refetch(prompt: str, type: str, fn: callable = None, *args, **kwargs):
    exception = None
    auto_select = kwargs.get("auto_select", None)
    not_found_message = kwargs.get("not_found_message", None)
    with Spinner(f"Fetching {type}", f"Fetched {type}"):
        try:
            items = fn(*args)
        except Exception as e:
            exception = e
    if exception is not None:
        rich.print(f"\n[red]Error fetching {type}: {exception}\n")
        return exception
    if len(items) == 0:
        if not_found_message:
            rich.print(f"\n[yellow]{not_found_message}\n")
        else:
            rich.print(f"\n[yellow]No valid {type} found\n")
        return None

    if auto_select and len(items) == 1 and items[0].lower() == auto_select.lower():
        return auto_select

    items.insert(0, REFETCH)

    passed_default = kwargs.get("default", None)
    passed_mandatory = kwargs.get("mandatory", False)

    rich.print("")
    result = fuzzy(
        prompt,
        items,
        default=get_default(passed_default, items),
        mandatory=passed_mandatory
    )
    rich.print("")

    while result == REFETCH:
        result = fuzzy_with_refetch(prompt, type, fn, *args, **kwargs)
    return result

def confirm(message:str, default:bool = False) -> bool:
    try:
        return inquirer.confirm(message, default=default, keybindings=default_bindings).execute()
    except KeyboardInterrupt:
        return abort()

def text(message:str, default:str|None = None, validator:callable = None, invalid_message:str|None = None, **kwargs) -> str:
    if validator and not invalid_message:
        invalid_message = "Invalid input"
    try:
        return inquirer.text(
            message,
            default=default or "",
            keybindings=default_bindings,
            validate=validator,
            invalid_message=invalid_message,
            **kwargs
        ).execute()
    except KeyboardInterrupt:
        return abort()

def password(message:str, default:str|None = None, validator:callable = None, invalid_message:str|None = None) -> str:
    try:
        return inquirer.secret(
            message,
            default=default or "",
            keybindings=default_bindings,
            validate=validator,
            invalid_message=invalid_message
        ).execute()
    except KeyboardInterrupt:
        return abort()

def file(message: str, start_path:Path|None = None, pattern:str|Pattern|None = None, default:str|None = None, allow_freeform=False, **kwargs) -> str|None:
    try:
        return FuzzyFile(message, start_path, allow_freeform=allow_freeform, max_height=8, border=True, style=STYLE, **kwargs).execute()
    except KeyboardInterrupt:
        return abort()

class FuzzyFile(inquirer.fuzzy):
    def __init__(self, message: str, initial_path: Path|None = None, allow_freeform = False,  *args, **kwargs):
        self.initial_path = initial_path or Path()
        self.current_path = Path(self.initial_path)
        self.allow_freeform = allow_freeform

        kwargs["keybindings"] = {
            **default_bindings,
            "answer": [
                {"key": os.sep},
                {"key": "enter"},
                {"key": "tab"},
                {"key": "right"}
            ],
            **kwargs.get("keybindings", {})
        }

        super().__init__(message, *args, **kwargs, choices=self._get_choices)

    def _get_prompt_message(self) -> List[tuple[str, str]]:
        pre_answer = ("class:instruction", f" {self.instruction} " if self.instruction else " ")
        result = str(nat_path(self.current_path, self.initial_path))

        if result:
            sep = " " if self._amark else ""
            return [
                ("class:answermark", self._amark),
                ("class:answered_question", f"{sep}{self._message} "),
                ("class:answer", f"{result}{os.sep if not self.status['answered'] else ''}"),
            ]
        else:
            sep = " " if self._qmark else ""
            return [
                ("class:answermark", self._amark),
                ("class:questionmark", self._qmark),
                ("class:question", f"{sep}{self._message}"),
                pre_answer
            ]

    def _handle_enter(self, event: KeyPressEvent) -> None:
        try:
            fake_document = FakeDocument(self.result_value)
            self._validator.validate(fake_document)  # type: ignore
            cc = self.content_control
            if self._multiselect:
                self.status["answered"] = True
                if not self.selected_choices:
                    self.status["result"] = [cc.selection["name"]]
                    event.app.exit(result=[cc.selection["value"]])
                else:
                    self.status["result"] = self.result_name
                    event.app.exit(result=self.result_value)
            else:
                res_value = cc.selection["value"]
                self.current_path /= res_value
                if self.current_path.is_dir():
                    self._update_choices()
                else:
                    self.status["answered"] = True
                    self.status["result"] = cc.selection["name"]
                    event.app.exit(result=str(nat_path(self.current_path, self.initial_path)))
        except ValidationError as e:
            self._set_error(str(e))
        except IndexError:
            self.status["answered"] = True
            res = self._get_current_text() if self.allow_freeform else None
            if self._multiselect:
                res = [res] if res is not None else []
            self.status["result"] = res
            event.app.exit(result=res)
        
    def _get_choices(self, _ = None):
        choices = os.listdir(self.current_path)
        choices.append("..")
        return choices

    def _update_choices(self):
        raw_choices = self._get_choices()
        cc = self.content_control
        cc.selected_choice_index = 0
        cc._raw_choices = raw_choices
        cc.choices = cc._get_choices(raw_choices, None)
        cc._safety_check()
        cc._format_choices()
        self._buffer.reset()

#--------------------------------------------------
# Spinner
#--------------------------------------------------

class Spinner:
    """Shows a spinner control while a task is running.\n
    The finished_message will not be printed if there was an exception and the failed_message is provided.
    """
    busy = False
    delay = 0.1

    def __init__(self, message="", finished_message="", failed_message=None, delay=None):
        self.message = message
        self.finished_message = finished_message
        self.failed_message = failed_message
        self.spinner_generator = itertools.cycle([ "▰▱▱▱", "▰▰▱▱", "▰▰▰▱", "▰▰▰▰", "▱▰▰▰", "▱▱▰▰", "▱▱▱▰", "▱▱▱▱" ])
        if delay and float(delay):
            self.delay = delay

    def get_message(self):
        return rich_str(f"[magenta]{next(self.spinner_generator)} {self.message}").strip()

    def get_clear(self, message:str):
        return '\b'* len(message.encode('utf-8'))

    def spinner_task(self):
        while self.busy:
            message = self.get_message()
            sys.stdout.write(message)
            sys.stdout.flush()
            time.sleep(self.delay)
            sys.stdout.write(self.get_clear(message))
            sys.stdout.flush()

    def __enter__(self):
        self.busy = True
        if jupyter.ipython:
            message = self.get_message()
            sys.stdout.write(message)
            sys.stdout.flush()
        else:
            threading.Thread(target=self.spinner_task).start()

    def __exit__(self, exception, value, tb):
        self.busy = False
        if exception is not None:
            if self.failed_message is not None:
                rich.print(f"\n\n[yellow]{self.failed_message} {value}")
                return True
            return False
        time.sleep(self.delay)
        message = self.get_message()
        sys.stdout.write(self.get_clear(message))
        if self.finished_message != "":
            final_message = f"[green]▰▰▰▰ {self.finished_message}"
            final_message += " " * (len(message) - len(final_message))
            rich.print(final_message)
        elif self.finished_message == "":
            sys.stdout.write(self.get_clear(message))
            sys.stdout.write(" "*len(message.encode('utf-8')))
            sys.stdout.write(self.get_clear(message))
            sys.stdout.flush()



import json
import queue
import threading
import time

import logging
import traceback
import uuid

from relationalai.util.keys import get_key

# Don't insert these as attrs since they get their own columns
FILTER_ATTRS = {
    "event",
    "span",
    "id",
    "parent_id",
    "start_time",
    "start_timestamp",
    "end_time",
    "end_timestamp",
    "elapsed",
}

class SnowflakeLogger(logging.Handler):
    """
    This is a logging handler that inserts spans directly into a Snowflake table.

    It uses a queue and a worker thread to buffer spans and then insert them in batches.
    """
    
    def __init__(self, snowflake_conn, sleep_interval_s=2):
        super().__init__()
        self.trace_id = uuid.uuid4()
        self.snowflake_conn = snowflake_conn
        self.queue = queue.Queue()
        self.is_shut_down = False
        self.sleep_interval_s = sleep_interval_s
        self.worker_thread = threading.Thread(target=self._consume_loop)
        self.worker_thread.start()
        self.open_spans = {} # only accessed from worker thread
        print('snowflake logger started with trace id', self.trace_id)
    
    def emit(self, record):
        if record.msg["event"] in ("span_start", "span_end"):
            self.queue.put(record.msg)
    
    def _consume_loop(self):
        while True:
            while not self.queue.empty():
                try:
                    batch = self._get_batch()
                    self._send_batch(batch)
                except Exception as e:
                    print('snowflake logger error:', e)
                    traceback.print_exc()
            if self.is_shut_down:
                return
            time.sleep(self.sleep_interval_s)
    
    def _get_batch(self):
        batch = []
        try:
            while True:
                item = self.queue.get_nowait()
                batch.append(item)
        except queue.Empty:
            pass
        return batch
    
    def _send_batch(self, batch):
        if not batch:
            return
        
        inserts = []
    
        for event in batch:
            if event["event"] == "span_start":
                span = event["span"]
                self.open_spans[str(span.id)] = span
            elif event["event"] == "span_end":
                span = self.open_spans.pop(event["id"])
                # assemble attributes
                filtered_attrs = {}
                for k, v in span.attrs.items():
                    if k not in FILTER_ATTRS:
                        filtered_attrs[k] = v
                for k, v in event["end_attrs"].items():
                    if k not in FILTER_ATTRS:
                        filtered_attrs[k] = v
                # get key
                span_key = get_key(span)
                # add to batch
                inserts.append({
                    "id": str(span.id),
                    "parent_id": None if span.parent is None else str(span.parent.id),
                    "trace_id": str(self.trace_id),
                    "key": str(span_key) if span_key is not None else None,
                    "type": span.type,
                    "start_ts": span.start_timestamp,
                    "finish_ts": span.end_timestamp,
                    "attrs": json.dumps(filtered_attrs, default=default_serialize),
                })
        # execute
        if inserts:
            self.snowflake_conn.cursor().executemany(
                """
                INSERT INTO spans_raw (id, parent_id, trace_id, type, key, start_ts, finish_ts, attrs)
                VALUES (%(id)s, %(parent_id)s, %(trace_id)s, %(type)s, %(key)s, %(start_ts)s, %(finish_ts)s, %(attrs)s)
                """,
                inserts,
            )

    
    def shut_down(self):
        self.is_shut_down = True
        print('snowflake logger: waiting to shut down...')
        self.worker_thread.join()
        print('snowflake logger: shut down')

def default_serialize(obj):
    return '<skipped>'
